/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* Author: Daniel Rypl */

#include "Econfig.h"
#include <stdlib.h>

#include "Eview.h"
#include "Egraphic.h"
#include "Egeomops.h"
#include "Ecolors.h"
#include "Eactset.h"
#include "Enurbs.h"

#define NURBSC3D_PRIVATE_HEADER
#include "Enurbsc3d.h"



/* ========================================================================= */
/* PRIVATE  */
/* ========================================================================= */

#define STRING_BUFF_SIZE 4096

static char string_rep_buffer[STRING_BUFF_SIZE];

/* ========================================================================= */
/* PUBLIC  */
/* ========================================================================= */

/* multiplicity corresponds to open knot vector */

GraphicObj *
CreateNURBSC3D(int degree, int spans, 
							 WCRec *points, float *weights, 
							 double *distinct_knots, short *multiplicity)
{
  ENURBSC3DWCRec *thedp;
  GraphicObj *g_p;
  int i, j, knots, pnts;
  
  thedp = (ENURBSC3DWCRec *)make_node(sizeof(ENURBSC3DWCRec));
  g_p = EGCreateGraphics(&nurbsc3d_graphic_methods, (caddr_t)thedp,
			 sizeof(ENURBSC3DWCRec));

  if((thedp -> multiplicity = (short *)calloc(spans + 1, sizeof(short))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D failed\n");
    exit(0);
  }

	knots = 0;
  for (i = 0; i <= spans; i++) {
		thedp->multiplicity[i] = multiplicity[i];
		knots += multiplicity[i];
	}

  if((thedp -> knotvector = (double *)calloc(knots, sizeof(double))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D failed\n");
    exit(0);
  }

	knots = 0;
  for (i = 0; i <= spans; i++) {
		for(j = 0; j < multiplicity[i]; j++)thedp->knotvector[knots++] = distinct_knots[i];
	}

	pnts = knots - degree - 1;

  if((thedp -> points = (WCRec *)calloc(pnts, sizeof(WCRec))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D failed\n");
    exit(0);
  }
  if((thedp -> weights = (float *)calloc(pnts, sizeof(float))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D failed\n");
    exit(0);
  }
  
  thedp -> degree = degree;
  thedp -> spans = spans;
  thedp -> knots = knots;
	thedp -> pnts = pnts;
  for (i = 0; i < pnts; i++, points++) {
		if(*weights <= 0.0 && spans > 1){
			fprintf(stderr, "Only positive weights accepted for NURBSC3D with multiple spans\n");
			exit(0);
		}
    thedp->points[i].x               = points -> x; 
    thedp->points[i].y               = points -> y;
    thedp->points[i].z               = points -> z;
    thedp->weights[i]                = *weights++;
  }

  thedp->attributes.spec.tessel_intervals = 32;
  thedp->attributes.spec.show_polygon     = YES;
  thedp->attributes.spec.show_curve       = YES;
  thedp->attributes.spec.style            = LineSolid;
  thedp->attributes.spec.width            = 0;
  thedp->attributes.spec.color            = 0x0;
  thedp->attributes.spec.have_any         = NO;
  return g_p;
}



static void 
Draw(EView *v_p, GraphicObj *p)
{
  ENURBSC3DWCRec *thedp;
  
  thedp = (ENURBSC3DWCRec *)(p->spec.data);
  
  EVNURBSC3DPrimitive(v_p, p, thedp->degree, thedp->pnts, thedp->spans,
											thedp->multiplicity, thedp->knotvector, 
											thedp->points, thedp->weights,
											thedp->attributes.spec.style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.width,
											thedp->attributes.spec.tessel_intervals,
											thedp->attributes.spec.show_polygon,
											DRAW_MODE_DRAW);
}



static void 
BoundingBoxVC(EView *v_p, GraphicObj *p, VCRec *ll, VCRec *ur)
{
  ENURBSC3DWCRec *thedp;
	BOOLEAN negative = NO, polygon = NO;
  VCRec pvc;
	WCRec pnt;
	double Bt[MAX_NURBS_DEGREE + 1], *bt = Bt, weight, tmp;
  int pnts, i, degree, order;
  
  thedp = (ENURBSC3DWCRec *)(p->spec.data);
  pnts = thedp->pnts;
  
  EVWCtoVC(v_p, &(thedp->points[0]), &pvc);
  ll->u = pvc.u;
  ll->v = pvc.v;
  ll->n = pvc.n;
  ur->u = pvc.u;
  ur->v = pvc.v;
  ur->n = pvc.n;

  EVWCtoVC(v_p, &(thedp->points[pnts - 1]), &pvc);
	ll->u = min(pvc.u, ll->u);
	ll->v = min(pvc.v, ll->v);
	ll->n = min(pvc.n, ll->n);
	ur->u = max(pvc.u, ur->u);
	ur->v = max(pvc.v, ur->v);
	ur->n = max(pvc.n, ur->n);

  for(i = 1; i < pnts - 1; i++){
		if(thedp->weights[i] < 0.0){
			negative = YES;
			break;
		}
  }

#ifdef PRECISE_BBOX_NURBS
	polygon = thedp->attributes.spec.show_polygon;

	if(polygon == NO || negative == YES){
		degree=thedp->degree;
		order = degree + 1;
		if(degree > MAX_NURBS_DEGREE)bt = (double *)calloc(order, sizeof(double));
		
		if(bt == NULL){
			fprintf(stderr, "Memory allocation for NURBSC3D in BoundingBoxVC failed\n");
			polygon = YES;
		}
		else{
			double t, dt, du, dv, dn;
			int j, k, count, cnt, sp, span, first, knot_p = 0, knot_n;
			
			count = 1 << (order - 2);
			if(negative == YES)count *= 2;
			cnt = count - 1;                                    /* skip last knot in each span (coincides with the first knot of subsequent span or with the last knot in the last span) */
			for(span = 0; span < thedp->spans; span++){         /* span within distinct knots */
				knot_n = knot_p + thedp->multiplicity[span];
				sp = knot_n - 1;                                  /* span within knot vector */
				first = GetFirstNonzeroBasisFunction(degree,sp);
				dt = (thedp->knotvector[knot_n] - thedp->knotvector[knot_p])/count;
				for (t = thedp->knotvector[knot_p], k = 0; k <= cnt; k++, t += dt) {
					if(span == 0 && k == 0)continue;                /* skip first knot in the first span */
					//sp = GetSpan(degree, pnts, thedp->knotvector, (double)t);
					BasisFunctions(degree, sp, thedp->knotvector, (double)t, bt);
					pnt.x = pnt.y = pnt.z = 0.0;
					weight = 0.0;
					i = first;
					for(j = 0; j < order; j++, i++){
						weight += (tmp = thedp->weights[i] * bt[j]);
						pnt.x += thedp->points[i].x * tmp;
						pnt.y += thedp->points[i].y * tmp;
						pnt.z += thedp->points[i].z * tmp;
					}
					pnt.x /= weight;
					pnt.y /= weight;
					pnt.z /= weight;
					
					//fprintf(stderr, "c: %f: %f %f %f\n", t, pnt.x, pnt.y, pnt.z);

					EVWCtoVC(v_p, &pnt, &pvc);
					ll->u = min(pvc.u, ll->u);
					ll->v = min(pvc.v, ll->v);
					ll->n = min(pvc.n, ll->n);
					ur->u = max(pvc.u, ur->u);
					ur->v = max(pvc.v, ur->v);
					ur->n = max(pvc.n, ur->n);

					//fprintf(stderr, "c: %f: %f %f %f\n", t, pvc.u, pvc.v, pvc.n);
				}
				knot_p = knot_n;
			}

			//fprintf(stderr, "c bb: %f %f %f %f %f %f\n", ll->u, ur->u, ll->v, ur->v, ll->n, ur->n);

#ifdef INCREASE_BBOX_NURBS

			/* increase the BBOX slightly */

			du = (ur->u - ll->u) / BBOX_PORTION_NURBS;
			dv = (ur->v - ll->v) / BBOX_PORTION_NURBS;
			dn = (ur->n - ll->n) / BBOX_PORTION_NURBS;
			ll->u -= du;
			ll->v -= dv;
			ll->n -= dn;
			ur->u += du;
			ur->v += dv;
			ur->n += dn;
#endif

			if(degree > MAX_NURBS_DEGREE)free(bt);
		}
	}
#else /* PRECISE_BBOX_NURBS */
	polygon = YES;

/* note: only very simple treatment of curve with negative weight is applied
         assuming that the curve is a rational bezier curve (nurbs curve with just a single nonzero knot span) !!! */

	if(negative == YES && thedp->spans == 1){
		degree=thedp->degree;
		order = degree + 1;
		if(degree > MAX_NURBS_DEGREE)bt = (double *)calloc(order, sizeof(double));

		if(bt == NULL)
			fprintf(stderr, "Memory allocation for NURBSC3D in BoundingBoxVC failed\n");
		else{
			double du, dv, dn;
			int sp;

			sp = degree;
			//sp = thedp->multiplicity[0] - 1;
			//sp = GetSpan(degree, pnts, thedp->knotvector, 0.5);
			BasisFunctions(degree, sp, thedp->knotvector, 0.5, bt);
			pnt.x = pnt.y = pnt.z = 0.0;
			weight = 0.0;
			for(i = 0; i < order; i++){
				weight += (tmp = thedp->weights[i] * bt[i]);
				pnt.x += thedp->points[i].x * tmp;
				pnt.y += thedp->points[i].y * tmp;
				pnt.z += thedp->points[i].z * tmp;
			}
			pnt.x /= weight;
			pnt.y /= weight;
			pnt.z /= weight;
			
			EVWCtoVC(v_p, &pnt, &pvc);
			ll->u = min(pvc.u, ll->u);
			ll->v = min(pvc.v, ll->v);
			ll->n = min(pvc.n, ll->n);
			ur->u = max(pvc.u, ur->u);
			ur->v = max(pvc.v, ur->v);
			ur->n = max(pvc.n, ur->n);

#ifdef INCREASE_BBOX_NURBS

			/* increase the BBOX slightly */

			du = (ur->u - ll->u) / BBOX_PORTION_NURBS;
			dv = (ur->v - ll->v) / BBOX_PORTION_NURBS;
			dn = (ur->n - ll->n) / BBOX_PORTION_NURBS;
			ll->u -= du;
			ll->v -= dv;
			ll->n -= dn;
			ur->u += du;
			ur->v += dv;
			ur->n += dn;
#endif

			if(degree > MAX_NURBS_DEGREE)free(bt);
		}
	}
#endif /* PRECISE_BBOX_NURBS */

	if(polygon == YES){

#ifdef INCREASE_BBOX_NURBS
		/* do no increase BB corresponding to ctrl points !!! */
#endif

		for(i = 1; i < pnts - 1; i++){
			EVWCtoVC(v_p, &(thedp->points[i]), &pvc);
			ll->u = min(pvc.u, ll->u);
			ll->v = min(pvc.v, ll->v);
			ll->n = min(pvc.n, ll->n);
			ur->u = max(pvc.u, ur->u);
			ur->v = max(pvc.v, ur->v);
			ur->n = max(pvc.n, ur->n);
		}
	}
}



static void 
BoundingBoxWC(GraphicObj *p, WCRec *ll, WCRec *ur)
{
  ENURBSC3DWCRec *thedp;
	BOOLEAN negative = NO, polygon = NO;
	WCRec pnt;
	double Bt[MAX_NURBS_DEGREE + 1], *bt = Bt, weight, tmp;
  int pnts, i, degree, order;

  thedp = (ENURBSC3DWCRec *)(p->spec.data);
  pnts = thedp->pnts;
  
  ll->x = thedp->points[0].x;
  ll->y = thedp->points[0].y;
  ll->z = thedp->points[0].z;
  ur->x = thedp->points[0].x;
  ur->y = thedp->points[0].y;
  ur->z = thedp->points[0].z;

	ll->x = min(ll->x, thedp->points[pnts - 1].x);
	ll->y = min(ll->y, thedp->points[pnts - 1].y);
	ll->z = min(ll->z, thedp->points[pnts - 1].z);
	ur->x = max(ur->x, thedp->points[pnts - 1].x); 
	ur->y = max(ur->y, thedp->points[pnts - 1].y);
	ur->z = max(ur->z, thedp->points[pnts - 1].z);

  for(i = 1; i < pnts - 1; i++){
		if(thedp->weights[i] < 0.0){
			negative = YES;
			break;
		}
  }

#ifdef PRECISE_BBOX_NURBS
	polygon = thedp->attributes.spec.show_polygon;

	if(polygon == NO || negative == YES){
		degree=thedp->degree;
		order = degree + 1;
		if(degree > MAX_NURBS_DEGREE)bt = (double *)calloc(order, sizeof(double));

		if(bt == NULL){
			fprintf(stderr, "Memory allocation for NURBSC3D in BoundingBoxWC failed\n");
			polygon = YES;
		}
		else{
			double t, dt, dx, dy, dz;
			int j, k, count, cnt, sp, span, first, knot_p = 0, knot_n;

			count = 1 << (order - 2);
			if(negative == YES)count *= 2;
			cnt--;                                              /* skip last knot in each span (coincides with the first knot of subsequent span or with the last knot in the last span) */
			for(span = 0; span < thedp->spans; span++){         /* span within distinct knots */
				knot_n = knot_p + thedp->multiplicity[span];
				sp = knot_n - 1;                                  /* span within knot vector */
				first = GetFirstNonzeroBasisFunction(degree,sp);
				dt = (thedp->knotvector[knot_n] - thedp->knotvector[knot_p])/count;
				for (t = thedp->knotvector[knot_p], k = 0; k <= cnt; k++, t += dt) {
					if(span == 0 && k == 0)continue;                /* skip first knot in the first span */
					//sp = GetSpan(degree, pnts, thedp->knotvector, (double)t);
					BasisFunctions(degree, sp, thedp->knotvector, (double)t, bt);
					pnt.x = pnt.y = pnt.z = 0.0;
					weight = 0.0;
					i = first;
					for(j = 0; j < order; j++, i++){
						weight += (tmp = thedp->weights[i] * bt[j]);
						pnt.x += thedp->points[i].x * tmp;
						pnt.y += thedp->points[i].y * tmp;
						pnt.z += thedp->points[i].z * tmp;
					}
					pnt.x /= weight;
					pnt.y /= weight;
					pnt.z /= weight;
					
					//fprintf(stderr, "c: %f: %f %f %f\n", t, pnt.x, pnt.y, pnt.z);

					ll->x = min(ll->x, pnt.x);
					ll->y = min(ll->y, pnt.y);
					ll->z = min(ll->z, pnt.z);
					ur->x = max(ur->x, pnt.x); 
					ur->y = max(ur->y, pnt.y);
					ur->z = max(ur->z, pnt.z);
				}
			}

#ifdef INCREASE_BBOX_NURBS

			/* increase the BBOX slightly */

			dx = (ur->x - ll->x) / BBOX_PORTION_NURBS;
			dy = (ur->y - ll->y) / BBOX_PORTION_NURBS;
			dz = (ur->z - ll->z) / BBOX_PORTION_NURBS;
			ll->x -= dx;
			ll->y -= dy;
			ll->z -= dz;
			ur->x += dx;
			ur->y += dy;
			ur->z += dz;
#endif

			if(degree > MAX_NURBS_DEGREE)free(bt);
		}
	}
#else /* PRECISE_BBOX_NURBS */
	polygon = YES;

/* note: only very simple treatment of curve with negative weight is applied
         assuming that the curve is a rational bezier curve (nurbs curve with just a single nonzero knot span) !!! */

	if(negative == YES && thedp->spans == 1){
		degree=thedp->degree;
		order = degree + 1;
		if(degree > MAX_NURBS_DEGREE)bt = (double *)calloc(order, sizeof(double));

		if(bt == NULL)
			fprintf(stderr, "Memory allocation for NURBSC3D in BoundingBoxWC failed\n");
		else{
			double dx, dy, dz;
			int sp;

			sp = degree;
			//sp = thedp->multiplicity[0] - 1;
			//sp = GetSpan(degree, pnts, thedp->knotvector, 0.5);
			BasisFunctions(degree, sp, thedp->knotvector, 0.5, bt);
			pnt.x = pnt.y = pnt.z = 0.0;
			weight = 0.0;
			for(i = 0; i < order; i++){
				weight += (tmp = thedp->weights[i] * bt[i]);
				pnt.x += thedp->points[i].x * tmp;
				pnt.y += thedp->points[i].y * tmp;
				pnt.z += thedp->points[i].z * tmp;
			}
			pnt.x /= weight;
			pnt.y /= weight;
			pnt.z /= weight;
			
			ll->x = min(ll->x, pnt.x);
			ll->y = min(ll->y, pnt.y);
			ll->z = min(ll->z, pnt.z);
			ur->x = max(ur->x, pnt.x); 
			ur->y = max(ur->y, pnt.y);
			ur->z = max(ur->z, pnt.z);

#ifdef INCREASE_BBOX_NURBS

			/* increase the BBOX slightly */

			dx = (ur->x - ll->x) / BBOX_PORTION_NURBS;
			dy = (ur->y - ll->y) / BBOX_PORTION_NURBS;
			dz = (ur->z - ll->z) / BBOX_PORTION_NURBS;
			ll->x -= dx;
			ll->y -= dy;
			ll->z -= dz;
			ur->x += dx;
			ur->y += dy;
			ur->z += dz;
#endif

			if(degree > MAX_NURBS_DEGREE)free(bt);
		}
	}
#endif /* PRECISE_BBOX_NURBS */

	if(polygon == YES){

#ifdef INCREASE_BBOX_NURBS
		/* do no increase BB corresponding to ctrl points !!! */
#endif

		for(i = 1; i < pnts - 1; i++){
			ll->x = min(ll->x, thedp->points[i].x);
			ll->y = min(ll->y, thedp->points[i].y);
			ll->z = min(ll->z, thedp->points[i].z);
			ur->x = max(ur->x, thedp->points[i].x); 
			ur->y = max(ur->y, thedp->points[i].y);
			ur->z = max(ur->z, thedp->points[i].z);
		}
	}
}



static void 
XorDraw(EView *v_p, GraphicObj *g_p)
{
  ENURBSC3DWCRec *thedp;
  
  thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  
  EVNURBSC3DPrimitive(v_p, g_p, thedp->degree, thedp->pnts, thedp->spans,
											thedp->multiplicity, thedp->knotvector, 
											thedp->points, thedp->weights,
											thedp->attributes.spec.style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.width,
											thedp->attributes.spec.tessel_intervals,
											thedp->attributes.spec.show_polygon,
											DRAW_MODE_XORDRAW);
}




static GraphicObj *
ModifyGeometry(GraphicObj *g_p, WCRec *points)
{
  ENURBSC3DWCRec *thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  int i, pnts = thedp->pnts;
  
  for (i = 0; i < pnts; i++, points++) {
    thedp->points[i].x               = points -> x; 
    thedp->points[i].y               = points -> y;
    thedp->points[i].z               = points -> z;
  }
  return g_p;
}




static BOOLEAN
GetGeometry(GraphicObj *g_p, WCRec *points, int *npoints)     /* array points must be large enough */
{
  ENURBSC3DWCRec *thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  int i, pnts = thedp->pnts;
  
  for (i = 0; i < pnts; i++, points++) {
    points -> x = thedp->points[i].x; 
    points -> y = thedp->points[i].y;
    points -> z = thedp->points[i].z;
  }
  *npoints = pnts;
  return YES;
}




static GraphicObj *
ModifyWeights(GraphicObj *g_p, FPNum *weights)
{
  ENURBSC3DWCRec *thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  int i, pnts = thedp->pnts;

  for (i = 0; i < pnts; i++)
    thedp->weights[i] = *weights++;
  return g_p;
}




static BOOLEAN
GetWeights(GraphicObj *g_p, FPNum *weights, int *nweights)   /* array weights must be large enough */
{
  ENURBSC3DWCRec *thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  int i, pnts = thedp->pnts;
  
  for (i = 0; i < pnts; i++)
    *weights++ = thedp->weights[i];
  *nweights = pnts;
  return YES;
}


static void 
Hilite(EView *v_p, GraphicObj *g_p)
{
  ENURBSC3DWCRec *thedp;
  
  thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  
  EVNURBSC3DPrimitive(v_p, g_p, thedp->degree, thedp->pnts, thedp->spans,
											thedp->multiplicity, thedp->knotvector, 
											thedp->points, thedp->weights,
											thedp->attributes.spec.style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.width,
											thedp->attributes.spec.tessel_intervals,
											thedp->attributes.spec.show_polygon,
											DRAW_MODE_HILITE);
}



static void 
Unhilite(EView *v_p, GraphicObj *g_p)
{
  ENURBSC3DWCRec *thedp;
  
  thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  
  EVNURBSC3DPrimitive(v_p, g_p, thedp->degree, thedp->pnts, thedp->spans,
											thedp->multiplicity, thedp->knotvector, 
											thedp->points, thedp->weights,
											thedp->attributes.spec.style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.width,
											thedp->attributes.spec.tessel_intervals,
											thedp->attributes.spec.show_polygon,
											DRAW_MODE_UNHILITE);
}



static void 
Erase(EView *v_p, GraphicObj *g_p)
{
  ENURBSC3DWCRec *thedp;
  
  thedp = (ENURBSC3DWCRec *)(g_p->spec.data);
  
  EVNURBSC3DPrimitive(v_p, g_p, thedp->degree, thedp->pnts, thedp->spans,
											thedp->multiplicity, thedp->knotvector, 
											thedp->points, thedp->weights,
											thedp->attributes.spec.style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.width,
											thedp->attributes.spec.tessel_intervals,
											thedp->attributes.spec.show_polygon,
											DRAW_MODE_ERASE); 
}



static GraphicObj *
Translate(GraphicObj *g_p, WCRec *by)
{
  ENURBSC3DWCRec *p;
  int i, pnts;
  
  p = (ENURBSC3DWCRec *)g_p->spec.data;
  pnts = p->pnts;
  for (i = 0; i < pnts; i++) {
    p->points[i].x += by->x;
    p->points[i].y += by->y;
    p->points[i].z += by->z;
  }
  return g_p;
}



static GraphicObj *
Rotate(GraphicObj *g_p, WCRec *center, WCRec *axial)
{
  ENURBSC3DWCRec *c;
  WCRec x_c;
  int i, pnts;
  
  c = (ENURBSC3DWCRec *)g_p->spec.data;
  pnts = c->pnts;
  for (i = 0; i < pnts; i++ ) {
    x_c.x = c->points[i].x - center->x;
    x_c.y = c->points[i].y - center->y;
    x_c.z = c->points[i].z - center->z;
    RotVectAboutVect(axial, &x_c);
    c->points[i].x = center->x + x_c.x;
    c->points[i].y = center->y + x_c.y;
    c->points[i].z = center->z + x_c.z;
  }
  return g_p;
}



static GraphicObj  *
MirrorPlane(GraphicObj *g_p,
            WCRec *point_on_plane, WCRec *unit_plane_normal)
{
  ENURBSC3DWCRec *dp;
  int j, pnts;
  
  dp = (ENURBSC3DWCRec *)g_p->spec.data;
  pnts = dp -> pnts;
  for (j = 0; j < pnts; j++) 
    MirrorPointInPlane(point_on_plane, unit_plane_normal, &dp->points[j]);
  
  return g_p;
}



static GraphicObj  *
MirrorCenter(GraphicObj *g_p, WCRec *center)
{
  ENURBSC3DWCRec *dp;
  int j, pnts;
  
  dp = (ENURBSC3DWCRec *)g_p->spec.data;
  pnts = dp -> pnts;
  for (j = 0; j < pnts; j++) {
    dp->points[j].x = 2.*center->x - dp->points[j].x;
    dp->points[j].y = 2.*center->y - dp->points[j].y;
    dp->points[j].z = 2.*center->z - dp->points[j].z;
  }
  return g_p;
}



static GraphicObj *
Scale(GraphicObj *g_p,
      WCRec *center, FPNum sx, FPNum sy, FPNum sz)
{
  return g_p;
}



static BOOLEAN 
HitByBox(EView *v_p, EBoxVC2DRec *box,
         ESelectCriteria sel_crit, GraphicObj *p)
{
  VCRec ll, ur, pvc;
  WC2DRec *points;
  ENURBSC3DWCRec *thedp;
  int i, pnts;
  BOOLEAN hit;
  
  switch(sel_crit) {
  case INSIDE:
    BoundingBoxVC(v_p, p, &ll, &ur);
    if (ll.u >= box->left   && ur.u <= box->right && 
				ll.v >= box->bottom && ur.v <= box->top)
      return YES;
    else
      return NO;
    break;
  case OVERLAP:
  case INTERSECT:
  default:
    thedp = (ENURBSC3DWCRec *)p->spec.data;
    pnts = thedp->pnts;
    if((points = (WC2DRec *)calloc(pnts, sizeof(WC2DRec))) == NULL){
      fprintf(stderr, "Memory allocation for NURBSC3D in HitByBox failed\n");
      return(NO);
    }
    for (i = 0; i < pnts; i++) {
      EVWCtoVC(v_p, &(thedp->points[i]), &pvc);
      points[i].x = pvc.u; points[i].y = pvc.v;
    }
    hit = BoxIntersectsNURBSCurve(box, thedp->degree, thedp->spans, thedp->knotvector, thedp->multiplicity, pnts, points, thedp->weights);
    free(points);
    return(hit);
    break;
  }
}



static EGraphicType 
Type(GraphicObj *g_p)
{
  return EG_NURBSCURVE3D;
}



static char *
AsString(GraphicObj *g_p)
{
  ENURBSC3DWCRec *l;
  int i, pnts, knot = 0, size, sz;
  char buff[128];
  
  l = (ENURBSC3DWCRec *)g_p->spec.data;
  pnts = l->pnts;
  sprintf(string_rep_buffer,
	  "NURBSCURVE3D %d \n", l->degree);
	size = (int)strlen(string_rep_buffer);
  for(i = 0; i < pnts; i++){
    sprintf(buff, "  [%f,%f,%f|%f] \n",
            l->points[i].x, l->points[i].y, l->points[i].z, l->weights[i]);
		sz = (int)strlen(buff);
		if((size += sz) > STRING_BUFF_SIZE - 20){
			sprintf(buff, "  . . . \n");
			strcat(string_rep_buffer, buff);
			return string_rep_buffer; 
		}
    strcat(string_rep_buffer, buff);
  }
	sprintf(buff, "  [");
	sz = (int)strlen(buff);
	if((size += sz) > STRING_BUFF_SIZE - 20){
		sprintf(buff, "  . . . \n");
		strcat(string_rep_buffer, buff);
		return string_rep_buffer; 
	}
	strcat(string_rep_buffer, buff);
	for(i = 0; i < l->spans; i++){
		if(l->multiplicity[i] == 1)
			sprintf(buff, "%f,", l->knotvector[knot]);
		else
			sprintf(buff, "%f(%dx),", l->knotvector[knot], l->multiplicity[i]);
		sz = (int)strlen(buff);
		if((size += sz) > STRING_BUFF_SIZE - 10){
			sprintf(buff, " . . . \n");
			strcat(string_rep_buffer, buff);
			return string_rep_buffer; 
		}
    strcat(string_rep_buffer, buff);
		knot += l->multiplicity[i];
	}
	i = l->spans;
	if(l->multiplicity[i] == 1)
		sprintf(buff, "%f]\n", l->knotvector[knot]);
	else
		sprintf(buff, "%f(%dx)]\n", l->knotvector[knot], l->multiplicity[i]);
	sz = (int)strlen(buff);
	if((size += sz) > STRING_BUFF_SIZE - 10){
		sprintf(buff, " . . . \n");
		strcat(string_rep_buffer, buff);
		return string_rep_buffer; 
	}
	strcat(string_rep_buffer, buff);
  return string_rep_buffer; /* Don't free this memory!!! */
}



static LIST 
KeyPoints(GraphicObj *p)
{
  LIST temp_list;
  KeyPointRec *key_p;
  ENURBSC3DWCRec *l_p;
  int pnts, i;
  
  l_p       = (ENURBSC3DWCRec *)(p->spec.data);
  pnts = l_p -> pnts;
  temp_list = make_list();
  
  for(i = 0; i < pnts; i++){
    key_p = (KeyPointRec*)add_to_tail(temp_list, make_node(sizeof(KeyPointRec)));
    key_p->coords.x = l_p->points[i].x;
    key_p->coords.y = l_p->points[i].y;
    key_p->coords.z = l_p->points[i].z;
    key_p->dist_from_hit = 0.0;
  }
  return temp_list;
}



static GraphicObj *
Reshape(GraphicObj *gp, EHandleNum hn, WCRec *pos)
{
  ENURBSC3DWCRec *_p;
  int i, pnts;
  
  _p = (ENURBSC3DWCRec *)(gp->spec.data);
  pnts = _p->pnts;
  
  if (hn < 1 || hn > pnts) 
    i = 0;
	else
		i = hn - 1;
  
  _p->points[i].x = pos->x;
  _p->points[i].y = pos->y;
  _p->points[i].z = pos->z;
  return gp;
}



static EHandleNum   
ReshapeHandle(EView *v_p, GraphicObj *p, WCRec *nearp)
{
  ENURBSC3DWCRec *l_p;
  VCRec nvc, pvc;
  FPNum d, dist;
  int i, pnts, hn;
  
  l_p       = (ENURBSC3DWCRec *)(p->spec.data);
  pnts = l_p -> pnts;
  
  EVWCtoVC(v_p, nearp, &nvc);
  EVWCtoVC(v_p, &(l_p->points[0]), &pvc);
  dist = DistOfPnts(pvc.u, pvc.v, nvc.u, nvc.v);
  hn = 0;
  for(i = 1; i < pnts; i++){
    EVWCtoVC(v_p, &(l_p->points[i]), &pvc);
    d = DistOfPnts(pvc.u, pvc.v, nvc.u, nvc.v);
    if(d < dist){
      dist = d;
      hn = i;
    }
  }
  return((EHandleNum)(hn + 1));
}



static void 
ShowHandles(EView *v_p, GraphicObj *p)
{
  ENURBSC3DWCRec *l_p;
  int i, pnts;
  
  l_p = (ENURBSC3DWCRec *)p->spec.data;
  pnts = l_p -> pnts;
  for(i = 0; i < pnts; i++)
    EGDrawHandle(v_p, &(l_p->points[i]));
}



static void 
EraseHandles(EView *v_p, GraphicObj *p)
{
  ENURBSC3DWCRec *l_p;
  int i, pnts;
  
  l_p = (ENURBSC3DWCRec *)p->spec.data;
  pnts = l_p -> pnts;
  for(i = 0; i < pnts; i++)
    EGEraseHandle(v_p, &(l_p->points[i]));
}




static void 
XYofHandle(GraphicObj *p, EHandleNum hn, WCRec *pos)
{
  ENURBSC3DWCRec *l_p;
  int i, pnts;
  
  l_p = (ENURBSC3DWCRec *)p->spec.data;
  pnts = l_p -> pnts;

  if (hn < 1 || hn > pnts)
    i = 0;
	else
		i = hn - 1;

  pos->x = l_p->points[i].x;
  pos->y = l_p->points[i].y;
  pos->z = l_p->points[i].z;
}



static void 
RetrieveFrom(FILE *file_p, GraphicSpecObj *gsp)
{
  ENURBSC3DWCRec *p;
  int style, width, sp, sc, i, j, ti, pnts, degree, spans, knots, multi;
  char color[64];
  EPixel pixel;
  BOOLEAN success;
  double x, y, z, weight, knot;
  
  if ((p = (ENURBSC3DWCRec *)make_node(sizeof(ENURBSC3DWCRec))) != NULL) {
    fscanf(file_p, "%d %d", &degree, &spans);

		if((p -> multiplicity = (short *)calloc(spans + 1, sizeof(short))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSC3D in RetrieveFrom failed\n");
			exit(0);
		}

		knots = 0;
    for (i = 0; i <= spans; i++) {
      fscanf(file_p, "%d", &multi);
			knots += (p->multiplicity[i] = multi);
		}

		if((p -> knotvector = (double *)calloc(knots, sizeof(double))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSC3D in RetrieveFrom failed\n");
			exit(0);
		}

		knots = 0;
    for (i = 0; i <= spans; i++) {
      fscanf(file_p, "%lf", &knot);
			for (j = 0; j < p->multiplicity[i]; j++)
				p->knotvector[knots++] = knot;
		}

		pnts = knots - degree - 1;

    p -> degree = degree;
    p -> spans = spans;
    p -> knots = knots;
    p -> pnts = pnts;
    if((p -> points = (WCRec *)calloc(pnts, sizeof(WCRec))) == NULL){
      fprintf(stderr, "Memory allocation for NURBSC3D in RetrieveFrom failed\n");
      exit(0);
    }
    if((p -> weights = (float *)calloc(pnts, sizeof(float))) == NULL){
      fprintf(stderr, "Memory allocation for NURBSC3D in RetrieveFrom failed\n");
      exit(0);
    }
    for (i = 0; i < pnts; i++) {
      fscanf(file_p, "%lf %lf %lf %lf", &x, &y, &z, &weight);
      p->points[i].x = x;
      p->points[i].y = y;
      p->points[i].z = z;
      p->weights[i] = weight;
    }
    fscanf(file_p,
	   "%d %d %d %d %d %d %s",
	   &(p->attributes.spec.have_any),
	   &sp, &sc, &width, &style, &ti, color);
    p->attributes.spec.show_polygon     = sp;
    p->attributes.spec.show_curve       = sc;
    p->attributes.spec.width            = width;
    p->attributes.spec.style            = style;
    p->attributes.spec.tessel_intervals = ti;
    pixel = ColorGetPixelFromString(color, &success);
    if (success)
      p->attributes.spec.color = pixel;
  }
  gsp->data        = (caddr_t)p;
  gsp->data_length = sizeof(ENURBSC3DWCRec);
}



static void 
StoreOn(FILE *file_p, GraphicObj *g_p)
{
  ENURBSC3DWCRec *p;
  int style, width, sp, sc, ti, i, pnts, knot = 0;
  char *color;
  
  p = (ENURBSC3DWCRec *)g_p->spec.data;
  pnts = p->pnts;
  if (p->attributes.spec.have_any == YES) {
    style = p->attributes.spec.style;
    width = p->attributes.spec.width;
    sp    = p->attributes.spec.show_polygon;
    sc    = p->attributes.spec.show_curve;
    ti    = p->attributes.spec.tessel_intervals;
    color = ColorGetStringFromPixel(p->attributes.spec.color);
    if (color == NULL)
      color = "black";
  } else {
    style = LineSolid;
    sp    = YES;
    sc    = YES;
    ti    = 31;
    width = 0;
    color = "black";
  }
  fprintf(file_p, "%d %d\n", p->degree, p->spans);
	for(i = 0; i <= p->spans; i++)
		fprintf(file_p, "%d ", p->multiplicity[i]);
  fprintf(file_p, "\n");
	for(i = 0; i <= p->spans; i++){
		fprintf(file_p, "%f ", p->knotvector[knot]);
		knot += p->multiplicity[i];
	}
  fprintf(file_p, "\n");
  for(i = 0; i < pnts; i++)
    fprintf(file_p, "%f %f %f %f \n",
            p->points[i].x, p->points[i].y, p->points[i].z, p->weights[i]);
  fprintf(file_p, "%d %d %d %d %d %d %s\n",	  p->attributes.spec.have_any,
	  sp, sc, width, style, ti, color);
}



static void 
ChangeAttributes(GraphicObj *p)
{
  ENURBSC3DWCRec *dp;
  unsigned long mask;
  
  mask = EASValsGetChangeMask();
  dp = (ENURBSC3DWCRec *)p->spec.data;
  if ((mask & SHOW_POLY_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.show_polygon = EASValsGetShowPoly();
    dp->attributes.spec.have_any     = YES;
  }
  if ((mask & SHOW_ENTITY_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.show_curve   = EASValsGetShowEntity();
    dp->attributes.spec.have_any     = YES;
  }
  if ((mask & TESSEL_INTERVALS_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.tessel_intervals = EASValsGetTesselIntervals();
    dp->attributes.spec.have_any         = YES;
  }
  if ((mask & STYLE_MASK) || mask & (ALL_ATTRIB_MASK)) {
    dp->attributes.spec.style    = EASValsGetLineStyle();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & WIDTH_MASK) || mask & (ALL_ATTRIB_MASK)) {
    dp->attributes.spec.width    = EASValsGetLineWidth();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & COLOR_MASK) || mask & (ALL_ATTRIB_MASK)) {
    dp->attributes.spec.color    = EASValsGetColor();
    dp->attributes.spec.have_any = YES;
  }
}



static void 
CopyAttributesToActiveSet(GraphicObj *p)
{
  ENURBSC3DWCRec *dp;
  
  dp = (ENURBSC3DWCRec *)p->spec.data;
  
  if (dp->attributes.spec.have_any) {
    EASValsSetLineStyle(dp->attributes.spec.style);
    EASValsSetLineWidth(dp->attributes.spec.width);
    EASValsSetColor(dp->attributes.spec.color);
    EASValsSetShowPoly(dp->attributes.spec.show_polygon);
    EASValsSetShowEntity(dp->attributes.spec.show_curve);
    EASValsSetTesselIntervals(dp->attributes.spec.tessel_intervals);
  }
}


static BOOLEAN
AttribValAsInAS(GraphicObj *p, unsigned long mask)  
{
  ENURBSC3DWCRec *dp;
  
  dp = (ENURBSC3DWCRec *)p->spec.data;
  
  if ((mask & SHOW_POLY_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.show_polygon != EASValsGetShowPoly())
      return NO;
  
  if ((mask & SHOW_ENTITY_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.show_curve   != EASValsGetShowEntity())
      return NO;
  
  if ((mask & TESSEL_INTERVALS_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.tessel_intervals != EASValsGetTesselIntervals())
      return NO;
  
  if ((mask & STYLE_MASK) || mask & (ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.style    != EASValsGetLineStyle())
      return NO;
  
  if ((mask & WIDTH_MASK) || mask & (ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.width    != EASValsGetLineWidth())
      return NO;
  
  if ((mask & COLOR_MASK) || mask & (ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.color    != EASValsGetColor())
      return NO;
  
  return YES;
}


static GraphicObj *
DeepCopy(GraphicObj *p)
{
  GraphicObj *gp;
  ENURBSC3DWCRec *thedp;
  caddr_t new_data_p;
  WCRec *points_p, *points_gp;
  float *weights_p, *weights_gp;
	double *knotvector_p, *knotvector_gp;
	short *multiplicity_p, *multiplicity_gp;
  int i, degree, pnts, spans, knots;
  
  /*	gp = EGDefaultCopyGraphics(p); */
  if ((new_data_p =
       (caddr_t)copy_node(p->spec.data, p->spec.data_length)) == NULL) {
    fprintf(stderr, "Failed copy data for NURBSC3D\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  gp = EGCreateGraphics(p->methods, new_data_p, p->spec.data_length);
  
  thedp = (ENURBSC3DWCRec *)(gp->spec.data);
  degree = thedp->degree;
  spans = thedp->spans;
  knots = thedp->knots;
  pnts = thedp->pnts;
	//knots = pnts + degree + 1;
  
  if((thedp -> multiplicity = (short *)calloc(spans + 1, sizeof(short))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if((thedp -> knotvector = (double *)calloc(knots, sizeof(double))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }

  if((thedp -> points = (WCRec *)calloc(pnts, sizeof(WCRec))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if((thedp -> weights = (float *)calloc(pnts, sizeof(float))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSC3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  
  multiplicity_p = ((ENURBSC3DWCRec *)(p->spec.data)) -> multiplicity;
  knotvector_p = ((ENURBSC3DWCRec *)(p->spec.data)) -> knotvector;
  points_p = ((ENURBSC3DWCRec *)(p->spec.data)) -> points;
  weights_p = ((ENURBSC3DWCRec *)(p->spec.data)) -> weights;
  multiplicity_gp = thedp -> multiplicity;
  knotvector_gp = thedp -> knotvector;
  points_gp = thedp -> points;
  weights_gp = thedp -> weights;
  
  for(i = 0; i <= spans; i++)multiplicity_gp[i] = multiplicity_p[i];
	for(i = 0; i < knots; i++)knotvector_gp[i] = knotvector_p[i];
  for(i = 0; i < pnts; i++){
    points_gp[i].x = points_p[i].x;
    points_gp[i].y = points_p[i].y;
    points_gp[i].z = points_p[i].z;
    weights_gp[i] = weights_p[i];
  }
  return(gp);
}



static void
DeepDestroy(GraphicObj *p)
{
  ENURBSC3DWCRec *thedp;
  thedp = (ENURBSC3DWCRec *)(p->spec.data);
  
  free(thedp->multiplicity);
  free(thedp->knotvector);
  free(thedp->points);
  free(thedp->weights);
  
  /* 	EGDefaultDestroyGraphics(p); */
  free_node(thedp);                        /* free specialized data */
  free_node(p);                            /* free top struct */
}


