/*

              ####### #         ###   #     #   ###   ######
              #       #          #     #   #     #    #     #
              #       #          #      # #      #    #     #
              #####   #          #       #       #    ######
              #       #          #      # #      #    #   #
              #       #          #     #   #     #    #    #
              ####### #######   ###   #     #   ###   #     #

                   
                  Copyright: 1994 Petr Krysl

   Czech Technical University in Prague, Faculty of Civil Engineering,
      Dept. Structural Mechanics, 166 29 Prague, Czech Republic,
                  email: pk@power2.fsv.cvut.cz
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

/* Author: Daniel Rypl */

#include "Econfig.h"
#include <stdlib.h>

#include "Eview.h"
#include "Egraphic.h"
#include "Egeomops.h"
#include "Ecolors.h"
#include "Eactset.h"
#include "Enurbs.h"

#define NURBSS3D_PRIVATE_HEADER
#include "Enurbss3d.h"



/* ========================================================================= */
/* PRIVATE  */
/* ========================================================================= */

#define STRING_BUFF_SIZE 8192

static char string_rep_buffer[STRING_BUFF_SIZE];

/* ========================================================================= */
/* PUBLIC  */
/* ========================================================================= */

/* multiplicity corresponds to open knot vector */

GraphicObj *
CreateNURBSS3D(int degree_u, int degree_v, int spans_u, int spans_v,
							 WCRec *points, float *weights, 
							 double *distinct_knots_u,  double *distinct_knots_v, 
							 short *multiplicity_u, short *multiplicity_v)
{
  ENURBSS3DWCRec *thedp;
  GraphicObj *g_p;
  int i, j, knots_u, knots_v, pnts_u, pnts_v, pnts;
  WCRec *p;
  float *w;
  
  thedp = (ENURBSS3DWCRec *)make_node(sizeof(ENURBSS3DWCRec));
  g_p = EGCreateGraphics(&nurbss3d_graphic_methods, (caddr_t)thedp,
			 sizeof(ENURBSS3DWCRec));

  if((thedp -> multiplicity_u = (short *)calloc(spans_u + 1, sizeof(short))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D failed\n");
    exit(0);
  }
  if((thedp -> multiplicity_v = (short *)calloc(spans_v + 1, sizeof(short))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D failed\n");
    exit(0);
  }

	knots_u = 0;
  for (i = 0; i <= spans_u; i++) {
		thedp->multiplicity_u[i] = multiplicity_u[i];
		knots_u += multiplicity_u[i];
	}
	knots_v = 0;
  for (i = 0; i <= spans_v; i++) {
		thedp->multiplicity_v[i] = multiplicity_v[i];
		knots_v += multiplicity_v[i];
	}

  if((thedp -> knotvector_u = (double *)calloc(knots_u, sizeof(double))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D failed\n");
    exit(0);
  }
  if((thedp -> knotvector_v = (double *)calloc(knots_v, sizeof(double))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D failed\n");
    exit(0);
  }

	knots_u = 0;
  for (i = 0; i <= spans_u; i++) {
		for(j = 0; j < multiplicity_u[i]; j++)thedp->knotvector_u[knots_u++] = distinct_knots_u[i];
	}
	knots_v = 0;
  for (i = 0; i <= spans_v; i++) {
		for(j = 0; j < multiplicity_v[i]; j++)thedp->knotvector_v[knots_v++] = distinct_knots_v[i];
	}

	pnts_u = knots_u - degree_u - 1;
	pnts_v = knots_v - degree_v - 1;
	pnts = pnts_u * pnts_v;

  if((thedp -> points = p
      = (WCRec *)calloc(pnts, sizeof(WCRec))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D failed\n");
    exit(0);
  }
  if((thedp -> weights = w
      = (float *)calloc(pnts, sizeof(float))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D failed\n");
    exit(0);
  }
  
  thedp->degree_u = degree_u;
  thedp->degree_v = degree_v;
  thedp->spans_u = spans_u;
  thedp->spans_v = spans_v;
  thedp->knots_u = knots_u;
  thedp->knots_v = knots_v;
  thedp->pnts_u = pnts_u;
  thedp->pnts_v = pnts_v;
  for(i = pnts; i > 0; i--, p++, points++){
		if(*weights <= 0.0 && spans_u > 1 && spans_v > 1){
			fprintf(stderr, "Only positive weights accepted for NURBSS3D with multiple spans in both directions\n");
			exit(0);
		}
    p->x = points->x;
    p->y = points->y;
    p->z = points->z;
    *w++ = *weights++;
  }
  thedp->attributes.spec.tessel_u         = 16;
  thedp->attributes.spec.tessel_v         = 16;
  thedp->attributes.spec.show_polygon     = YES;
  thedp->attributes.spec.show_surface     = YES;
  thedp->attributes.spec.line_style       = LineSolid;
  thedp->attributes.spec.line_width       = 0;
  thedp->attributes.spec.fill_style       = FILL_HOLLOW;
  thedp->attributes.spec.edge_flag        = NO;
  thedp->attributes.spec.edge_color       = 0x0;
  thedp->attributes.spec.color            = 0x0;
  thedp->attributes.spec.have_any         = NO;
  return g_p;
}



static void 
Draw(EView *v_p, GraphicObj *p)
{
  ENURBSS3DWCRec *thedp;
  
  thedp = (ENURBSS3DWCRec *)(p->spec.data);
  
  EVNURBSS3DPrimitive(v_p, p, thedp->degree_u, thedp->degree_v, 
											thedp->pnts_u, thedp->pnts_v,
											thedp->spans_u, thedp->spans_v,
											thedp->multiplicity_u, thedp->multiplicity_v, 
											thedp->knotvector_u, thedp->knotvector_v, 
											thedp->points, thedp->weights,
											thedp->attributes.spec.line_style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.line_width,
											thedp->attributes.spec.fill_style,
											thedp->attributes.spec.edge_flag,
											thedp->attributes.spec.edge_color,
											thedp->attributes.spec.tessel_u,
											thedp->attributes.spec.tessel_v,
											thedp->attributes.spec.show_polygon,
											NO,
											DRAW_MODE_DRAW);
}



static void 
BoundingBoxVC(EView *v_p, GraphicObj *p, VCRec *ll, VCRec *ur)
{
  ENURBSS3DWCRec *thedp;
	BOOLEAN polygon = NO, negative = NO;
  VCRec pvc;
	WCRec pnt;
	double Bu[MAX_NURBS_DEGREE + 1], *bu = Bu, Bv[MAX_NURBS_DEGREE + 1], *bv = Bv, weight, tmp;
  int i, j, k, pnts, degree_u, degree_v, order_u, order_v;
  
  thedp = (ENURBSS3DWCRec *)(p->spec.data);
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  
  EVWCtoVC(v_p, &(thedp->points[0]), &pvc);
  ll->u = pvc.u;
  ll->v = pvc.v;
  ll->n = pvc.n;
  ur->u = pvc.u;
  ur->v = pvc.v;
  ur->n = pvc.n;

  for(i = 1; i < pnts - 1; i++){
		if(thedp->weights[i] < 0.0){
			negative = YES;
			break;
		}
	}

#ifdef PRECISE_BBOX_NURBS
	polygon = thedp->attributes.spec.show_polygon;

	if(polygon == NO || negative == YES){
		degree_u=thedp->degree_u;
		degree_v=thedp->degree_v;
		order_u = degree_u + 1;
		order_v = degree_v + 1;
		if(degree_u > MAX_NURBS_DEGREE)bu = (double *)calloc(order_u, sizeof(double));
		if(bu == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
			polygon = YES;
			negative = NO;
		}
		else{
			if(degree_v > MAX_NURBS_DEGREE)bv = (double *)calloc(order_v, sizeof(double));
			if(bv == NULL){
				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
				polygon = YES;
				negative = NO;
			}
			else{
				double u, v, du, dv, dn;
				int count_u, cnt_u, sp_u, span_u, first_u, knot_u_p, knot_u_n;
				int count_v, cnt_v, sp_v, span_v, first_v, knot_v_p, knot_v_n;
				int n, m;

/* order   1<<(order-2)   order-1   order-1+[(1<<(order-2)-1)/order]
    2           1            1          1
    3           2            2          2
    4           4            3          3
    5           8            4          5
    6          16            5          7
    7          32            6         10

1<<(order-2)                       slow, precise
order-1                            fast, inacurate
order-1+(1<<(order-2)-1/order      reasonable compromise for positive weights
(order-1)*3                        reasonable compromise for negative weights for single span surface (in particular direction) */

				if(negative == NO){
					count_u = 1 << (order_u - 2);
					count_v = 1 << (order_v - 2);

					count_u = order_u - 1 + (int)((count_u - 1)/order_u);
					count_v = order_v - 1 + (int)((count_v - 1)/order_v);
				}
				else{
					count_u = (order_u - 1) * 3;
					count_v = (order_v - 1) * 3;
				}
		
				knot_u_p = 0;
				for(span_u = 0; span_u < thedp->spans_u; span_u++){         /* span within distinct knots */
					knot_u_n = knot_u_p + thedp->multiplicity_u[span_u];
					sp_u = knot_u_n - 1;                                      /* span within knot vector */
					first_u = GetFirstNonzeroBasisFunction(degree_u,sp_u);
					du = (thedp->knotvector_u[knot_u_n] - thedp->knotvector_u[knot_u_p])/count_u;
					if(span_u == thedp->spans_u - 1)
						cnt_u = count_u;
					else
						cnt_u = count_u - 1;                                    /* skip last knot (coincides with the first knot of subsequent span) */

					knot_v_p = 0;
					for(span_v = 0; span_v < thedp->spans_v; span_v++){       /* span within distinct knots */
						knot_v_n = knot_v_p + thedp->multiplicity_v[span_v];
						sp_v = knot_v_n - 1;                                    /* span within knot vector */
						first_v = GetFirstNonzeroBasisFunction(degree_v,sp_v);
						dv = (thedp->knotvector_v[knot_v_n] - thedp->knotvector_v[knot_v_p])/count_v;
						if(span_v == thedp->spans_v - 1)
							cnt_v = count_v;
						else
							cnt_v = count_v - 1;                                  /* skip last knot (coincides with the first knot of subsequent span) */

						for (u = thedp->knotvector_u[knot_u_p], n = 0; n <= cnt_u; n++, u += du) {
							if(span_u == thedp->spans_u - 1 && n == count_u)u = thedp->knotvector_v[knot_u_n];
							//sp_u = GetSpan(degree_u, thedp->pnts_u, thedp->knotvector_u, (double)u);
							BasisFunctions(degree_u, sp_u, thedp->knotvector_u, (double)u, bu);

							for (v = thedp->knotvector_v[knot_v_p], m = 0; m <= cnt_v; m++, v += dv) {
								if(span_v == thedp->spans_v - 1 && m == count_v)v = thedp->knotvector_v[knot_v_n];
								//sp_v = GetSpan(degree_v, thedp->pnts_v, thedp->knotvector_v, (double)v);
								BasisFunctions(degree_v, sp_v, thedp->knotvector_v, (double)v, bv);

								weight = 0.0;
								pnt.x = pnt.y = pnt.z = 0.0;
								k = first_v * thedp->pnts_u + first_u;
								for(j = 0; j < order_v; j++){
									for(i = 0; i < order_u; i++, k++){
										weight += (tmp = thedp->weights[k] * bu[i] * bv[j]);
										pnt.x += thedp->points[k].x * tmp;
										pnt.y += thedp->points[k].y * tmp;
										pnt.z += thedp->points[k].z * tmp;
									}
									k += thedp->pnts_u - order_u;
								}
								pnt.x /= weight;
								pnt.y /= weight;
								pnt.z /= weight;

								//fprintf(stderr, "s: %f %f: %f %f %f\n", u, v, pnt.x, pnt.y, pnt.z);
				
								EVWCtoVC(v_p, &pnt, &pvc);
								ll->u = min(pvc.u, ll->u);
								ll->v = min(pvc.v, ll->v);
								ll->n = min(pvc.n, ll->n);
								ur->u = max(pvc.u, ur->u);
								ur->v = max(pvc.v, ur->v);
								ur->n = max(pvc.n, ur->n);

								//fprintf(stderr, "s: %f %f: %f %f %f\n", u, v, pvc.u, pvc.v, pvc.n);
							}
						}
						knot_v_p = knot_v_n;
					}
					knot_u_p = knot_u_n;
				}

				//fprintf(stderr, "s bb: %f %f %f %f %f %f\n", ll->u, ur->u, ll->v, ur->v, ll->n, ur->n);

#ifdef INCREASE_BBOX_NURBS

				/* increase the BBOX slightly */

				du = (ur->u - ll->u) / BBOX_PORTION_NURBS;
				dv = (ur->v - ll->v) / BBOX_PORTION_NURBS;
				dn = (ur->n - ll->n) / BBOX_PORTION_NURBS;
				ll->u -= du;
				ll->v -= dv;
				ll->n -= dn;
				ur->u += du;
				ur->v += dv;
				ur->n += dn;
#endif

				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				if(degree_v > MAX_NURBS_DEGREE)free(bv);
			}
		}
	}
#else /* PRECISE_BBOX_NURBS */
	polygon = YES;

/* note: only very simple treatment of surface with negative weight is applied
         assuming that the surface is a rational bezier surface (nurbs surface with just a single nonzero knot span) !!! */
  
	if(negative == YES && thedp->spans_u == 1 && thedp->spans_v == 1){
		degree_u=thedp->degree_u;
		degree_v=thedp->degree_v;
		order_u = degree_u + 1;
		order_v = degree_v + 1;
		if(degree_u > MAX_NURBS_DEGREE)bu = (double *)calloc(order_u, sizeof(double));
		if(bu == NULL)
			fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
		else{
			if(degree_v > MAX_NURBS_DEGREE)bv = (double *)calloc(order_v, sizeof(double));
			if(bv == NULL){
				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
			}
			else{
				double du, dv, dn;
				int sp_u, sp_v;

#ifdef INCREASE_BBOX_NURBS

				/* initiate BB to corner ctrl points */

				EVWCtoVC(v_p, &(thedp->points[thedp -> pnts_u - 1]), &pvc);
				ll->u = min(pvc.u, ll->u);
				ll->v = min(pvc.v, ll->v);
				ll->n = min(pvc.n, ll->n);
				ur->u = max(pvc.u, ur->u);
				ur->v = max(pvc.v, ur->v);
				ur->n = max(pvc.n, ur->n);

				EVWCtoVC(v_p, &(thedp->points[pnts - thedp -> pnts_u]), &pvc);
				ll->u = min(pvc.u, ll->u);
				ll->v = min(pvc.v, ll->v);
				ll->n = min(pvc.n, ll->n);
				ur->u = max(pvc.u, ur->u);
				ur->v = max(pvc.v, ur->v);
				ur->n = max(pvc.n, ur->n);

				EVWCtoVC(v_p, &(thedp->points[pnts - 1]), &pvc);
				ll->u = min(pvc.u, ll->u);
				ll->v = min(pvc.v, ll->v);
				ll->n = min(pvc.n, ll->n);
				ur->u = max(pvc.u, ur->u);
				ur->v = max(pvc.v, ur->v);
				ur->n = max(pvc.n, ur->n);
#endif

				sp_u = degree_u;
				//sp_u = thedp->multiplicity_u[0] - 1;
				//sp_u = GetSpan(degree_u, thedp->pnts_u, thedp->knotvector_u, 0.5);
				BasisFunctions(degree_u, sp_u, thedp->knotvector_u, 0.5, bu);

				sp_v = degree_v;
				//sp_v = thedp->multiplicity_v[0] - 1;
				//sp_v = GetSpan(degree_v, thedp->pnts_v, thedp->knotvector_v, 0.5);
				BasisFunctions(degree_v, sp_v, thedp->knotvector_v, 0.5, bv);

				weight = 0.0;
				pnt.x = pnt.y = pnt.z = 0.0;
				k = 0;
				for(j = 0; j < order_v; j++){
					for(i = 0; i < order_u; i++){
						weight += (tmp = thedp->weights[k] * bu[i] * bv[j]);
						pnt.x += thedp->points[k].x * tmp;
						pnt.y += thedp->points[k].y * tmp;
						pnt.z += thedp->points[k].z * tmp;
						k++;
					}
				}
				pnt.x /= weight;
				pnt.y /= weight;
				pnt.z /= weight;
				
				EVWCtoVC(v_p, &pnt, &pvc);
				ll->u = min(pvc.u, ll->u);
				ll->v = min(pvc.v, ll->v);
				ll->n = min(pvc.n, ll->n);
				ur->u = max(pvc.u, ur->u);
				ur->v = max(pvc.v, ur->v);
				ur->n = max(pvc.n, ur->n);

#ifdef INCREASE_BBOX_NURBS

				/* increase the BBOX slightly */

				du = (ur->u - ll->u) / BBOX_PORTION_NURBS;
				dv = (ur->v - ll->v) / BBOX_PORTION_NURBS;
				dn = (ur->n - ll->n) / BBOX_PORTION_NURBS;
				ll->u -= du;
				ll->v -= dv;
				ll->n -= dn;
				ur->u += du;
				ur->v += dv;
				ur->n += dn;
#endif

				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				if(degree_v > MAX_NURBS_DEGREE)free(bv);
			}
		}
	}
#endif  /* PRECISE_BBOX_NURBS */

	if(polygon == YES){

#ifdef INCREASE_BBOX_NURBS
		/* do no increase BB corresponding to ctrl points !!! */
#endif

		for(i = 1; i < pnts; i++){
			EVWCtoVC(v_p, &(thedp->points[i]), &pvc);
			ll->u = min(pvc.u, ll->u);
			ll->v = min(pvc.v, ll->v);
			ll->n = min(pvc.n, ll->n);
			ur->u = max(pvc.u, ur->u);
			ur->v = max(pvc.v, ur->v);
			ur->n = max(pvc.n, ur->n);
		}
	}
}



static void 
BoundingBoxWC(GraphicObj *p, WCRec *ll, WCRec *ur)
{
  ENURBSS3DWCRec *thedp;
	BOOLEAN polygon = NO, negative = NO;
	WCRec pnt, *ctrl_pnt = NULL;
	double Bu[MAX_NURBS_DEGREE + 1], *bu = Bu, Bv[MAX_NURBS_DEGREE + 1], *bv = Bv, weight, tmp;
  int i, j, k, pnts, degree_u, degree_v, order_u, order_v;
  
  thedp = (ENURBSS3DWCRec *)(p->spec.data);
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  
  ll->x = thedp->points[0].x;
  ll->y = thedp->points[0].y;
  ll->z = thedp->points[0].z;
  ur->x = thedp->points[0].x;
  ur->y = thedp->points[0].y;
  ur->z = thedp->points[0].z;

  for(i = 1; i < pnts - 1; i++){
		if(thedp->weights[i] < 0.0){
			negative = YES;
			break;
		}
	}


#ifdef PRECISE_BBOX_NURBS
	polygon = thedp->attributes.spec.show_polygon;

	if(polygon == NO || negative == YES){
		degree_u=thedp->degree_u;
		degree_v=thedp->degree_v;
		order_u = degree_u + 1;
		order_v = degree_v + 1;
		if(degree_u > MAX_NURBS_DEGREE)bu = (double *)calloc(order_u, sizeof(double));
		if(bu == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
			polygon = YES;
			negative = NO;
		}
		else{
			if(degree_v > MAX_NURBS_DEGREE)bv = (double *)calloc(order_v, sizeof(double));
			if(bv == NULL){
				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
				polygon = YES;
				negative = NO;
			}
			else{
				double u, v, du, dv, dx, dy, dz;
				int count_u, cnt_u, sp_u, span_u, first_u, knot_u_p, knot_u_n;
				int count_v, cnt_v, sp_v, span_v, first_v, knot_v_p, knot_v_n;
				int n, m;

/* order   1<<(order-2)   order-1   order-1+[(1<<(order-2)-1)/order]
    2           1            1          1
    3           2            2          2
    4           4            3          3
    5           8            4          5
    6          16            5          7
    7          32            6         10

1<<(order-2)                       slow, precise
order-1                            fast, inacurate
order-1+(1<<(order-2)-1/order      reasonable compromise for positive weights
(order-1)*3                        reasonable compromise for negative weights for single span surface (in particular direction) */

				if(negative == NO){
					count_u = 1 << (order_u - 2);
					count_v = 1 << (order_v - 2);

					count_u = order_u - 1 + (int)((count_u - 1)/order_u);
					count_v = order_v - 1 + (int)((count_v - 1)/order_v);
				}
				else{
					count_u = (order_u - 1) * 3;
					count_v = (order_v - 1) * 3;
				}
		
				knot_u_p = 0;
				for(span_u = 0; span_u < thedp->spans_u; span_u++){         /* span within distinct knots */
					knot_u_n = knot_u_p + thedp->multiplicity_u[span_u];
					sp_u = knot_u_n - 1;                                      /* span within knot vector */
					first_u = GetFirstNonzeroBasisFunction(degree_u,sp_u);
					du = (thedp->knotvector_u[knot_u_n] - thedp->knotvector_u[knot_u_p])/count_u;
					if(span_u == thedp->spans_u - 1)
						cnt_u = count_u;
					else
						cnt_u = count_u - 1;                                    /* skip last knot (coincides with the first knot of subsequent span) */

					knot_v_p = 0;
					for(span_v = 0; span_v < thedp->spans_v; span_v++){       /* span within distinct knots */
						knot_v_n = knot_v_p + thedp->multiplicity_v[span_v];
						sp_v = knot_v_n - 1;                                    /* span within knot vector */
						first_v = GetFirstNonzeroBasisFunction(degree_v,sp_v);
						dv = (thedp->knotvector_v[knot_v_n] - thedp->knotvector_v[knot_v_p])/count_v;
						if(span_v == thedp->spans_v - 1)
							cnt_v = count_v;
						else
							cnt_v = count_v - 1;                                  /* skip last knot (coincides with the first knot of subsequent span) */

						for (u = thedp->knotvector_u[knot_u_p], n = 0; n <= cnt_u; n++, u += du) {
							if(span_u == thedp->spans_u - 1 && n == count_u - 1)u = thedp->knotvector_v[knot_u_n];
							//sp_u = GetSpan(degree_u, thedp->pnts_u, thedp->knotvector_u, (double)u);
							BasisFunctions(degree_u, sp_u, thedp->knotvector_u, (double)u, bu);

							for (v = thedp->knotvector_v[knot_v_p], m = 0; m <= cnt_v; m++, v += dv) {
								if(span_v == thedp->spans_v - 1 && m == count_v - 1)v = thedp->knotvector_v[knot_v_n];
								//sp_v = GetSpan(degree_v, thedp->pnts_v, thedp->knotvector_v, (double)v);
								BasisFunctions(degree_v, sp_v, thedp->knotvector_v, (double)v, bv);

								weight = 0.0;
								pnt.x = pnt.y = pnt.z = 0.0;
								k = first_v * thedp->pnts_u + first_u;
								for(j = 0; j < order_v; j++){
									for(i = 0; i < order_u; i++, k++){
										weight += (tmp = thedp->weights[k] * bu[i] * bv[j]);
										pnt.x += thedp->points[k].x * tmp;
										pnt.y += thedp->points[k].y * tmp;
										pnt.z += thedp->points[k].z * tmp;
									}
									k += thedp->pnts_u - order_u;
								}
								pnt.x /= weight;
								pnt.y /= weight;
								pnt.z /= weight;
				
								//fprintf(stderr, "s: %f %f: %f %f %f\n", u, v, pnt.x, pnt.y, pnt.z);
				
								ll->x = min(ll->x, pnt.x);
								ll->y = min(ll->y, pnt.y);
								ll->z = min(ll->z, pnt.z);
								ur->x = max(ur->x, pnt.x); 
								ur->y = max(ur->y, pnt.y);
								ur->z = max(ur->z, pnt.z);
							}
						}
						knot_v_p = knot_v_n;
					}
					knot_u_p = knot_u_n;
				}

#ifdef INCREASE_BBOX_NURBS

				/* increase the BBOX slightly */

				dx = (ur->x - ll->x) / BBOX_PORTION_NURBS;
				dy = (ur->y - ll->y) / BBOX_PORTION_NURBS;
				dz = (ur->z - ll->z) / BBOX_PORTION_NURBS;
				ll->x -= dx;
				ll->y -= dy;
				ll->z -= dz;
				ur->x += dx;
				ur->y += dy;
				ur->z += dz;
#endif

				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				if(degree_v > MAX_NURBS_DEGREE)free(bv);
			}
		}
	}
#else /* PRECISE_BBOX_NURBS */
	polygon = YES;

/* note: only very simple treatment of surface with negative weight is applied
         assuming that the surface is a rational bezier surface (nurbs surface with just a single nonzero knot span) !!! */
  
	if(negative == YES && thedp->spans_u == 1 && thedp->spans_v == 1){
		degree_u=thedp->degree_u;
		degree_v=thedp->degree_v;
		order_u = degree_u + 1;
		order_v = degree_v + 1;
		if(degree_u > MAX_NURBS_DEGREE)bu = (double *)calloc(order_u, sizeof(double));
		if(bu == NULL)
			fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
		else{
			if(degree_v > MAX_NURBS_DEGREE)bv = (double *)calloc(order_v, sizeof(double));
			if(bv == NULL){
				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				fprintf(stderr, "Memory allocation for NURBSS3D in BoundingBoxVC failed\n");
			}
			else{
				double dx, dy, dz;
				int sp_u, sp_v;

#ifdef INCREASE_BBOX_NURBS

				/* initiate BB to corner ctrl points */

				ctrl_pnt = &(thedp->points[order_u - 1]);
				ll->x = min(ll->x, ctrl_pnt->x);
				ll->y = min(ll->y, ctrl_pnt->y);
				ll->z = min(ll->z, ctrl_pnt->z);
				ur->x = max(ur->x, ctrl_pnt->x); 
				ur->y = max(ur->y, ctrl_pnt->y);
				ur->z = max(ur->z, ctrl_pnt->z);

				ctrl_pnt = &(thedp->points[order_u * order_v - order_u]);
				ll->x = min(ll->x, ctrl_pnt->x);
				ll->y = min(ll->y, ctrl_pnt->y);
				ll->z = min(ll->z, ctrl_pnt->z);
				ur->x = max(ur->x, ctrl_pnt->x); 
				ur->y = max(ur->y, ctrl_pnt->y);
				ur->z = max(ur->z, ctrl_pnt->z);

				ctrl_pnt = &(thedp->points[order_u * order_v - 1]);
				ll->x = min(ll->x, ctrl_pnt->x);
				ll->y = min(ll->y, ctrl_pnt->y);
				ll->z = min(ll->z, ctrl_pnt->z);
				ur->x = max(ur->x, ctrl_pnt->x); 
				ur->y = max(ur->y, ctrl_pnt->y);
				ur->z = max(ur->z, ctrl_pnt->z);
#endif

				sp_u = degree_u;
				//sp_u = thedp->multiplicity_u[0] - 1;
				//sp_u = GetSpan(degree_u, thedp->pnts_u, thedp->knotvector_u, 0.5);
				BasisFunctions(degree_u, sp_u, thedp->knotvector_u, 0.5, bu);

				sp_v = degree_v;
				//sp_v = thedp->multiplicity_v[0] - 1;
				//sp_v = GetSpan(degree_v, thedp->pnts_v, thedp->knotvector_v, 0.5);
				BasisFunctions(degree_v, sp_v, thedp->knotvector_v, 0.5, bv);

				weight = 0.0;
				pnt.x = pnt.y = pnt.z = 0.0;
				k = 0;
				for(j = 0; j < order_v; j++){
					for(i = 0; i < order_u; i++){
						weight += (tmp = thedp->weights[k] * bu[i] * bv[j]);
						pnt.x += thedp->points[k].x * tmp;
						pnt.y += thedp->points[k].y * tmp;
						pnt.z += thedp->points[k].z * tmp;
						k++;
					}
				}
				pnt.x /= weight;
				pnt.y /= weight;
				pnt.z /= weight;
				
				ll->x = min(ll->x, pnt.x);
				ll->y = min(ll->y, pnt.y);
				ll->z = min(ll->z, pnt.z);
				ur->x = max(ur->x, pnt.x); 
				ur->y = max(ur->y, pnt.y);
				ur->z = max(ur->z, pnt.z);

#ifdef INCREASE_BBOX_NURBS

				/* increase the BBOX slightly */
				
				dx = (ur->x - ll->x) / BBOX_PORTION_NURBS;
				dy = (ur->y - ll->y) / BBOX_PORTION_NURBS;
				dz = (ur->z - ll->z) / BBOX_PORTION_NURBS;
				ll->x -= dx;
				ll->y -= dy;
				ll->z -= dz;
				ur->x += dx;
				ur->y += dy;
				ur->z += dz;
#endif

				if(degree_u > MAX_NURBS_DEGREE)free(bu);
				if(degree_v > MAX_NURBS_DEGREE)free(bv);
			}
		}
	}
#endif  /* PRECISE_BBOX_NURBS */

	if(polygon == YES){

#ifdef INCREASE_BBOX_NURBS
		/* do no increase BB corresponding to ctrl points !!! */
#endif

		for(i = 1; i < pnts; i++){
			ll->x = min(ll->x, thedp->points[i].x);
			ll->y = min(ll->y, thedp->points[i].y);
			ll->z = min(ll->z, thedp->points[i].z);
			ur->x = max(ur->x, thedp->points[i].x); 
			ur->y = max(ur->y, thedp->points[i].y);
			ur->z = max(ur->z, thedp->points[i].z);
		}
	}
}



static void 
XorDraw(EView *v_p, GraphicObj *g_p)
{
  ENURBSS3DWCRec *thedp;
  
  thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  EVNURBSS3DPrimitive(v_p,  g_p, thedp->degree_u, thedp->degree_v, 
											thedp->pnts_u, thedp->pnts_v,
											thedp->spans_u, thedp->spans_v,
											thedp->multiplicity_u, thedp->multiplicity_v, 
											thedp->knotvector_u, thedp->knotvector_v, 
											thedp->points, thedp->weights, 
											thedp->attributes.spec.line_style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.line_width,
											FILL_HOLLOW, NO, 0,
											thedp->attributes.spec.tessel_u,
											thedp->attributes.spec.tessel_v,
											thedp->attributes.spec.show_polygon,
											NO,
											DRAW_MODE_XORDRAW);
}




static GraphicObj *
ModifyGeometry(GraphicObj *g_p, WCRec *points)
{
  ENURBSS3DWCRec *thedp;
  WCRec *p;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  p = thedp -> points;
  
  for (i = 0; i < pnts; i++, p++, points++){
    p->x = points->x;
    p->y = points->y;
    p->z = points->z;
  }
  return g_p;
}


static BOOLEAN 
GetGeometry(GraphicObj *g_p, WCRec *points, int *npoints)
     /* array points must be large enough */
{
  ENURBSS3DWCRec *thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  WCRec *p;
  int i, pnts;
  
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  p = thedp -> points;
  
  for (i = 0; i < pnts; i++, p++, points++){
    points->x = p->x;
    points->y = p->y;
    points->z = p->z;
  }
  *npoints = pnts;
  return YES;
}



static GraphicObj *
ModifyWeights(GraphicObj *g_p, FPNum *weights)
{
  ENURBSS3DWCRec *thedp;
  float *w;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  pnts = thedp->pnts_u * thedp -> pnts_v;
  w = thedp -> weights;
  
  for (i = 0; i < pnts; i++)
    *w++ = *weights++;
  return g_p;
}




static BOOLEAN
GetWeights(GraphicObj *g_p, FPNum *weights, int *nweights)
     /* array weights must be large enough */
{
  ENURBSS3DWCRec *thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  float *w;
  int i, pnts;
  
  pnts = thedp->pnts_u * thedp -> pnts_v;
  w = thedp -> weights;
  
  for (i = 0; i < pnts; i++)
    *weights++ = *w++;
  *nweights = pnts;
  return YES;
}



static void 
Hilite(EView *v_p, GraphicObj *g_p)
{
  ENURBSS3DWCRec *thedp;
  
  thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  EVNURBSS3DPrimitive(v_p,  g_p, thedp->degree_u, thedp->degree_v, 
											thedp->pnts_u, thedp->pnts_v,
											thedp->spans_u, thedp->spans_v,
											thedp->multiplicity_u, thedp->multiplicity_v, 
											thedp->knotvector_u, thedp->knotvector_v, 
											thedp->points, thedp -> weights,
											thedp->attributes.spec.line_style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.line_width,
											FILL_HOLLOW, NO, 0,
											thedp -> degree_u + 1, thedp -> degree_v + 1,
											thedp->attributes.spec.show_polygon,
											NO,
											DRAW_MODE_HILITE);
}



static void 
Unhilite(EView *v_p, GraphicObj *g_p)
{
  ENURBSS3DWCRec *thedp;
  
  thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  EVNURBSS3DPrimitive(v_p,  g_p, thedp->degree_u, thedp->degree_v, 
											thedp->pnts_u, thedp->pnts_v,
											thedp->spans_u, thedp->spans_v,
											thedp->multiplicity_u, thedp->multiplicity_v, 
											thedp->knotvector_u, thedp->knotvector_v, 
											thedp->points, thedp -> weights, 
											thedp->attributes.spec.line_style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.line_width,
											FILL_HOLLOW, NO, 0,
											thedp -> degree_u + 1, thedp -> degree_v + 1,
											thedp->attributes.spec.show_polygon,
											NO,
											DRAW_MODE_UNHILITE);
}



static void 
Erase(EView *v_p, GraphicObj *g_p)
{
  ENURBSS3DWCRec *thedp;
  
  thedp = (ENURBSS3DWCRec *)(g_p->spec.data);
  
  EVNURBSS3DPrimitive(v_p,  g_p, thedp->degree_u, thedp->degree_v, 
											thedp->pnts_u, thedp->pnts_v,
											thedp->spans_u, thedp->spans_v,
											thedp->multiplicity_u, thedp->multiplicity_v, 
											thedp->knotvector_u, thedp->knotvector_v, 
											thedp->points, thedp -> weights, 
											thedp->attributes.spec.line_style,
											thedp->attributes.spec.color,
											thedp->attributes.spec.line_width,
											thedp->attributes.spec.fill_style, NO, 0,
											thedp->attributes.spec.tessel_u,
											thedp->attributes.spec.tessel_v,
											thedp->attributes.spec.show_polygon,
											NO,
											DRAW_MODE_ERASE);  
}



static GraphicObj *
Translate(GraphicObj *g_p, WCRec *by)
{
  ENURBSS3DWCRec *thedp;
  WCRec *p;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)g_p->spec.data;
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  p = thedp -> points;
  
  for (i = 0; i < pnts; i++, p++){
    p->x += by->x;
    p->y += by->y;
    p->z += by->z;
  }
  return g_p;
}



static GraphicObj *
Rotate(GraphicObj *g_p, WCRec *center, WCRec *axial)
{
  ENURBSS3DWCRec *thedp;
  WCRec x_c, *p;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)g_p->spec.data;
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  p = thedp -> points;
  
  for (i = 0; i < pnts; i++, p++) {
    x_c.x = p->x - center->x;
    x_c.y = p->y - center->y;
    x_c.z = p->z - center->z;
    RotVectAboutVect(axial, &x_c);
    p->x = center->x + x_c.x;
    p->y = center->y + x_c.y;
    p->z = center->z + x_c.z;
  }
  return g_p;
}



static GraphicObj  *
MirrorPlane(GraphicObj *g_p,
            WCRec *point_on_plane, WCRec *unit_plane_normal)
{
  ENURBSS3DWCRec *thedp;
  WCRec *p;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)g_p->spec.data;
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  p = thedp -> points;
  
  for (i = 0; i < pnts; i++, p++)
    MirrorPointInPlane(point_on_plane, unit_plane_normal, p);
  
  return g_p;
}



static GraphicObj  *
MirrorCenter(GraphicObj *g_p, WCRec *center)
{
  ENURBSS3DWCRec *thedp;
  WCRec *p;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)g_p->spec.data;
  pnts = thedp -> pnts_u * thedp -> pnts_v;
  p = thedp -> points;
  
  for (i = 0; i < pnts; i++, p++){
    p->x = 2.*center->x - p->x;
    p->y = 2.*center->y - p->y;
    p->z = 2.*center->z - p->z;
  }
  return g_p;
}



static GraphicObj *
Scale(GraphicObj *g_p,
      WCRec *center, FPNum sx, FPNum sy, FPNum sz)
{
  return g_p;
}


static BOOLEAN 
HitByBox(EView *v_p, EBoxVC2DRec *box,
         ESelectCriteria sel_crit, GraphicObj *g_p)
{
  VCRec ll, ur, pvc;
  WCRec *p;
  WC2DRec *points = NULL;
  float *weights = NULL, *w;
  ENURBSS3DWCRec *thedp;
  int i, pnts_u, pnts_v, pnts;
  
  switch(sel_crit) {
  case INSIDE: /* only if the whole bounding box is inside */
    BoundingBoxVC(v_p, g_p, &ll, &ur);
    if (ll.u >= box->left   && ur.u <= box->right && ll.v >= box->bottom && ur.v <= box->top)
      return YES;
    else
      return NO;
    break;
  case OVERLAP:
  case INTERSECT: /* hits the boundary curves */
  default:
    thedp = (ENURBSS3DWCRec *)g_p->spec.data;
    pnts_u = thedp -> pnts_u;
    pnts_v = thedp -> pnts_v;
    p = thedp -> points;
    w = thedp -> weights;
		pnts = max(pnts_u, pnts_v);
		if((points = (WC2DRec *)calloc(pnts, sizeof(WC2DRec))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in HitByBox failed\n");
			return(NO);
		}
		if((weights = (float *)calloc(pnts, sizeof(float))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in HitByBox failed\n");
			free(points);
			return(NO);
		}
    for (i = 0; i < pnts_u; i++) {   /* curve 1 */
      EVWCtoVC(v_p, &p[i], &pvc);
      points[i].x = pvc.u; points[i].y = pvc.v;
      weights[i] = w[i];
    }
    if (BoxIntersectsNURBSCurve(box, thedp->degree_u, thedp->spans_u, thedp->knotvector_u, thedp->multiplicity_u, pnts_u, points, weights)){
			free(points);
			free(weights);
      return YES;
    }
    for (i = 0; i < pnts_v; i++) {      /* curve 2 */
      EVWCtoVC(v_p, &p[i * pnts_u], &pvc);
      points[i].x = pvc.u; points[i].y = pvc.v;
      weights[i] = w[i * pnts_u];
    }
    if (BoxIntersectsNURBSCurve(box, thedp->degree_v, thedp->spans_v, thedp->knotvector_v, thedp->multiplicity_v, pnts_v, points, weights)){
			free(points);
			free(weights);
      return YES;
    }
    for (i = 0; i < pnts_u; i++) {      /* curve 3 */
      EVWCtoVC(v_p, &p[i + (pnts_v - 1) * pnts_u], &pvc);
      points[i].x = pvc.u; points[i].y = pvc.v;
      weights[i] = w[i + (pnts_v - 1) * pnts_u];
    }
    if (BoxIntersectsNURBSCurve(box, thedp->degree_u, thedp->spans_u, thedp->knotvector_u, thedp->multiplicity_u, pnts_u, points, weights)){
			free(points);
			free(weights);
      return YES;
    }
    for (i = 0; i < pnts_v; i++) {     /* curve 4 */
      EVWCtoVC(v_p, &p[(i + 1) * pnts_u - 1], &pvc);
      points[i].x = pvc.u; points[i].y = pvc.v;
      weights[i] = w[(i + 1) * pnts_u - 1];
    }
    if (BoxIntersectsNURBSCurve(box, thedp->degree_v, thedp->spans_v, thedp->knotvector_v, thedp->multiplicity_v, pnts_v, points, weights)){
			free(points);
			free(weights);
      return YES;
    }
		free(points);
		free(weights);
    return NO;
    break;
  }
}



static EGraphicType 
Type(GraphicObj *g_p)
{
  return EG_NURBSSURFACE3D;
}



static char *
AsString(GraphicObj *g_p)
{
  ENURBSS3DWCRec *l;
  WCRec *p;
  float *w;
  int i, pnts, knot_u=0, knot_v=0, size, sz;
  char buff[128];
  
  l = (ENURBSS3DWCRec *)g_p->spec.data;
  pnts = l->pnts_u * l->pnts_v;
  p = l->points;
  w = l->weights;
  sprintf(string_rep_buffer,
	  "NURBSSURFACE3D %d %d \n", l->degree_u, l->degree_v);
	size = (int)strlen(string_rep_buffer);
  for(i = 0; i < pnts; i++){
    sprintf(buff, "  [%f,%f,%f|%f] \n",
            l->points[i].x, l->points[i].y, l->points[i].z, l->weights[i]);
		sz = (int)strlen(buff);
		if((size += sz) > STRING_BUFF_SIZE - 20){
			sprintf(buff, "  . . . \n");
			strcat(string_rep_buffer, buff);
			return string_rep_buffer; 
		}
    strcat(string_rep_buffer, buff);
  }
	sprintf(buff, "  [");
	sz = (int)strlen(buff);
	if((size += sz) > STRING_BUFF_SIZE - 20){
		sprintf(buff, "  . . . \n");
		strcat(string_rep_buffer, buff);
		return string_rep_buffer; 
	}
	strcat(string_rep_buffer, buff);
	for(i = 0; i < l->spans_u; i++){
		if(l->multiplicity_u[i] == 1)
			sprintf(buff, "%f,", l->knotvector_u[knot_u]);
		else
			sprintf(buff, "%f(%dx),", l->knotvector_u[knot_u], l->multiplicity_u[i]);
		sz = (int)strlen(buff);
		if((size += sz) > STRING_BUFF_SIZE - 10){
			sprintf(buff, " . . . \n");
			strcat(string_rep_buffer, buff);
			return string_rep_buffer; 
		}
    strcat(string_rep_buffer, buff);
		knot_u += l->multiplicity_u[i];
	}
	i = l->spans_u;
	if(l->multiplicity_u[i] == 1)
		sprintf(buff, "%f]\n", l->knotvector_u[knot_u]);
	else
		sprintf(buff, "%f(%dx)]\n", l->knotvector_u[knot_u], l->multiplicity_u[i]);
	sz = (int)strlen(buff);
	if((size += sz) > STRING_BUFF_SIZE - 10){
		sprintf(buff, " . . . \n");
		strcat(string_rep_buffer, buff);
		return string_rep_buffer; 
	}
	sprintf(buff, "  [");
	sz = (int)strlen(buff);
	if((size += sz) > STRING_BUFF_SIZE - 20){
		sprintf(buff, "  . . . \n");
		strcat(string_rep_buffer, buff);
		return string_rep_buffer; 
	}
	strcat(string_rep_buffer, buff);
	for(i = 0; i < l->spans_v; i++){
		if(l->multiplicity_v[i] == 1)
			sprintf(buff, "%f,", l->knotvector_v[knot_v]);
		else
			sprintf(buff, "%f(%dx),", l->knotvector_v[knot_v], l->multiplicity_v[i]);
		sz = (int)strlen(buff);
		if((size += sz) > STRING_BUFF_SIZE - 10){
			sprintf(buff, " . . . \n");
			strcat(string_rep_buffer, buff);
			return string_rep_buffer; 
		}
    strcat(string_rep_buffer, buff);
		knot_v += l->multiplicity_v[i];
	}
	i = l->spans_v;
	if(l->multiplicity_v[i] == 1)
		sprintf(buff, "%f]\n", l->knotvector_v[knot_v]);
	else
		sprintf(buff, "%f(%dx)]\n", l->knotvector_v[knot_v], l->multiplicity_v[i]);
	sz = (int)strlen(buff);
	if((size += sz) > STRING_BUFF_SIZE - 10){
		sprintf(buff, " . . . \n");
		strcat(string_rep_buffer, buff);
		return string_rep_buffer; 
	}
	strcat(string_rep_buffer, buff);
  return string_rep_buffer; /* Don't free this memory!!! */
}


static LIST 
KeyPoints(GraphicObj *g_p)
{
  LIST temp_list;
  KeyPointRec *key_p;
  ENURBSS3DWCRec *thedp;
  WCRec *p;
  int i, pnts;
  
  thedp       = (ENURBSS3DWCRec *)(g_p->spec.data);
  pnts = thedp->pnts_u * thedp->pnts_v;
  p = thedp->points;
  temp_list = make_list();
  
  for (i = 0; i < pnts; i++, p++){
    key_p = (KeyPointRec*)add_to_tail(temp_list,
                                      make_node(sizeof(KeyPointRec)));
    key_p->coords.x = p->x;
    key_p->coords.y = p->y;
    key_p->coords.z = p->z;
    key_p->dist_from_hit = 0.0;
  }
  return temp_list;
}



static GraphicObj *
Reshape(GraphicObj *gp, EHandleNum hn, WCRec *pos)
{
  ENURBSS3DWCRec *thedp;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)(gp->spec.data);
  pnts = thedp->pnts_u * thedp->pnts_v;
  
  i = hn - 1;
  if (hn < 1 || hn > pnts) 
    i = 0;
  
  thedp->points[i].x = pos->x;
  thedp->points[i].y = pos->y;
  thedp->points[i].z = pos->z;
  return gp;
}



static EHandleNum   
ReshapeHandle(EView *v_p, GraphicObj *gp, WCRec *nearp)
{
  ENURBSS3DWCRec *thedp;
  VCRec nvc, pvc;
  WCRec *p;
  EHandleNum hn;
  double dist, d;
  int i, pnts;
  
  thedp       = (ENURBSS3DWCRec *)(gp->spec.data);
  pnts = thedp->pnts_u * thedp->pnts_v;
  p = thedp->points;
  
  EVWCtoVC(v_p, nearp, &nvc);
  EVWCtoVC(v_p, p, &pvc);
  dist = DistOfPnts(pvc.u, pvc.v, nvc.u, nvc.v);
  hn = 0;
  for(i = 1, p++; i < pnts; i++, p++){
    EVWCtoVC(v_p, p, &pvc);
    d = DistOfPnts(pvc.u, pvc.v, nvc.u, nvc.v);
    if(d < dist){
      dist = d;
      hn = i;
    }
  }
  return((EHandleNum)(hn + 1));
}



static void 
ShowHandles(EView *v_p, GraphicObj *gp)
{
  ENURBSS3DWCRec *thedp;
  WCRec *p;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)gp->spec.data;
  pnts = thedp->pnts_u * thedp->pnts_v;
  p = thedp->points;
  
  for (i = 0; i < pnts; i++, p++)
    EGDrawHandle(v_p, p);
}



static void 
EraseHandles(EView *v_p, GraphicObj *gp)
{
  ENURBSS3DWCRec *thedp;
  WCRec *p;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)gp->spec.data;
  pnts = thedp->pnts_u * thedp->pnts_v;
  p = thedp->points;
  
  for (i = 0; i < pnts; i++, p++)
    EGEraseHandle(v_p, p);
}




static void 
XYofHandle(GraphicObj *p, EHandleNum hn, WCRec *pos)
{
  ENURBSS3DWCRec *thedp;
  int i, pnts;
  
  thedp = (ENURBSS3DWCRec *)p->spec.data;
  pnts = thedp->pnts_u * thedp->pnts_v;
  
  i = hn - 1;
  if (hn < 1 || hn > pnts)
    i = 0;
  pos->x = thedp->points[i].x;
  pos->y = thedp->points[i].y;
  pos->z = thedp->points[i].z;
}



static void 
RetrieveFrom(FILE *file_p, GraphicSpecObj  *gsp)
{
  ENURBSS3DWCRec *p;
  int fill, eflag, sp, sc, tu, tv, i, j, style, width, pnts, multi;
	int degree_u, degree_v, spans_u, spans_v, knots_u, knots_v, pnts_u, pnts_v;
  char color[64];
  EPixel pixel;
  BOOLEAN success;
  double x, y, z, weight, knot;
  
  if ((p = (ENURBSS3DWCRec *)make_node(sizeof(ENURBSS3DWCRec))) != NULL) {
    fscanf(file_p, "%d %d %d %d", &degree_u, &degree_v, &spans_u, &spans_v);

		if((p -> multiplicity_u = (short *)calloc(spans_u + 1, sizeof(short))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in RetrieveFrom failed\n");
			exit(0);
		}
		if((p -> multiplicity_v = (short *)calloc(spans_v + 1, sizeof(short))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in RetrieveFrom failed\n");
			exit(0);
		}

		knots_u = 0;
    for (i = 0; i <= spans_u; i++) {
      fscanf(file_p, "%d", &multi);
			knots_u += (p->multiplicity_u[i] = multi);
		}
		knots_v = 0;
    for (i = 0; i <= spans_v; i++) {
      fscanf(file_p, "%d", &multi);
			knots_v += (p->multiplicity_v[i] = multi);
		}

		if((p -> knotvector_u = (double *)calloc(knots_u, sizeof(double))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in RetrieveFrom failed\n");
			exit(0);
		}
		if((p -> knotvector_v = (double *)calloc(knots_v, sizeof(double))) == NULL){
			fprintf(stderr, "Memory allocation for NURBSS3D in RetrieveFrom failed\n");
			exit(0);
		}

		knots_u = 0;
    for (i = 0; i <= spans_u; i++) {
      fscanf(file_p, "%lf", &knot);
			for (j = 0; j < p->multiplicity_u[i]; j++)
				p->knotvector_u[knots_u++] = knot;
		}
		knots_v = 0;
    for (i = 0; i <= spans_v; i++) {
      fscanf(file_p, "%lf", &knot);
			for (j = 0; j < p->multiplicity_v[i]; j++)
				p->knotvector_v[knots_v++] = knot;
		}

		pnts_u = knots_u - degree_u - 1;
		pnts_v = knots_v - degree_v - 1;
		pnts = pnts_u * pnts_v;


    p -> degree_u = degree_u;
    p -> degree_v = degree_v;
    p -> spans_u = spans_u;
    p -> spans_v = spans_v;
    p -> knots_u = knots_u;
    p -> knots_v = knots_v;
    p -> pnts_u = pnts_u;
    p -> pnts_v = pnts_v;
    if((p -> points = (WCRec *)calloc(pnts, sizeof(WCRec)))
       == NULL){
      fprintf(stderr, "Memory allocation for NURBSS3D in RetrieveFrom failed\n");
      exit(0);
    }
    if((p -> weights = (float *)calloc(pnts, sizeof(float)))
       == NULL){
      fprintf(stderr, "Memory allocation for NURBSS3D in RetrieveFrom failed\n");
      exit(0);
    }
    for (i = 0; i < pnts; i++) {
      fscanf(file_p, "%lf %lf %lf %lf", &x, &y, &z, &weight);
      p->points[i].x = x;
      p->points[i].y = y;
      p->points[i].z = z;
      p->weights[i] = weight;
    }
    fscanf(file_p, "%d %d %d %d %d %d %d %d %d",
	   &(p->attributes.spec.have_any),
					 &sp, &sc, &fill, &eflag, &width, &style, &tu, &tv);
    p->attributes.spec.show_polygon     = sp;
    p->attributes.spec.show_surface     = sc;
    p->attributes.spec.fill_style       = fill;
    p->attributes.spec.edge_flag        = eflag;
    p->attributes.spec.line_width       = width;
    p->attributes.spec.line_style       = style;
    p->attributes.spec.tessel_u         = tu;
    p->attributes.spec.tessel_v         = tv;
    fscanf(file_p, "%s", color);
    pixel = ColorGetPixelFromString(color, &success);
    if (success)
      p->attributes.spec.color = pixel;
    fscanf(file_p, "%s", color);
    pixel = ColorGetPixelFromString(color, &success);
    if (success)
      p->attributes.spec.edge_color = pixel;
  }
  gsp->data        = (caddr_t)p;
  gsp->data_length = sizeof(ENURBSS3DWCRec);
}



static void 
StoreOn(FILE *file_p, GraphicObj *g_p)
{
  ENURBSS3DWCRec *p;
  int fill, sp, sc, tu, tv, eflag, i, style, width, pnts, knot_u = 0, knot_v = 0;
  char *color, *edge_color;
  
  p = (ENURBSS3DWCRec *)g_p->spec.data;
  pnts = p->pnts_u * p->pnts_v;
  if (p->attributes.spec.have_any == YES) {
    fill  = p->attributes.spec.fill_style;
    eflag = p->attributes.spec.edge_flag;
    sp    = p->attributes.spec.show_polygon;
    sc    = p->attributes.spec.show_surface;
    style = p->attributes.spec.line_style;
    width = p->attributes.spec.line_width;
    tu    = p->attributes.spec.tessel_u;
    tv    = p->attributes.spec.tessel_v;
    color = ColorGetStringFromPixel(p->attributes.spec.color);
    if (color == NULL)
      color = "black";
    edge_color = ColorGetStringFromPixel(p->attributes.spec.edge_color);
    if (edge_color == NULL)
      edge_color = "white";
  } else {
    fill  = FILL_HOLLOW;
    eflag = 0;
    sp    = YES;
    sc    = YES;
		style = LineSolid;
		width = 0;
    tu    = 16;
    tv    = 16;
    color = edge_color = "black";
  }
  fprintf(file_p, "%d %d %d %d\n", p->degree_u, p->degree_v, p->spans_u, p->spans_v);
	for(i = 0; i <= p->spans_u; i++)
		fprintf(file_p, "%d ", p->multiplicity_u[i]);
  fprintf(file_p, "\n");
	for(i = 0; i <= p->spans_v; i++)
		fprintf(file_p, "%d ", p->multiplicity_v[i]);
  fprintf(file_p, "\n");
	for(i = 0; i <= p->spans_u; i++){
		fprintf(file_p, "%f ", p->knotvector_u[knot_u]);
		knot_u += p->multiplicity_u[i];
	}
  fprintf(file_p, "\n");
	for(i = 0; i <= p->spans_v; i++){
		fprintf(file_p, "%f ", p->knotvector_v[knot_v]);
		knot_v += p->multiplicity_v[i];
	}
  fprintf(file_p, "\n");
  for(i = 0; i < pnts; i++)
    fprintf(file_p, "%f %f %f %f \n",
            p->points[i].x, p->points[i].y, p->points[i].z, p->weights[i]);
  fprintf(file_p, "%d %d %d %d %d %d %d %d %d\n%s\n%s\n",
	  p->attributes.spec.have_any,
					sp, sc, fill, eflag, width, style, tu, tv, color, edge_color);
}



static void 
ChangeAttributes(GraphicObj *p)
{
  ENURBSS3DWCRec *dp;
  unsigned long mask;
  
  mask = EASValsGetChangeMask();
  dp = (ENURBSS3DWCRec *)p->spec.data;
  if ((mask & SHOW_POLY_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.show_polygon = EASValsGetShowPoly();
    dp->attributes.spec.have_any     = YES;
  }
  if ((mask & SHOW_ENTITY_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.show_surface   = EASValsGetShowEntity();
    dp->attributes.spec.have_any     = YES;
  }
  if ((mask & TESSEL_U_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.tessel_u = EASValsGetUTesselIntervals();
    dp->attributes.spec.have_any         = YES;
  }
  if ((mask & TESSEL_V_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.tessel_v = EASValsGetVTesselIntervals();
    dp->attributes.spec.have_any         = YES;
  }
  if ((mask & FILL_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.fill_style   = EASValsGetFillStyle();
    dp->attributes.spec.have_any     = YES;
  }
  if ((mask & COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.color    = EASValsGetColor();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & EDGE_COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.edge_color    = EASValsGetEdgeColor();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & EDGE_FLAG_MASK) || (mask & ALL_ATTRIB_MASK)) {
    dp->attributes.spec.edge_flag    = EASValsGetEdgeFlag();
    dp->attributes.spec.have_any = YES; 
  }
  if ((mask & STYLE_MASK) || mask & (ALL_ATTRIB_MASK)) {
    dp->attributes.spec.line_style    = EASValsGetLineStyle();
    dp->attributes.spec.have_any = YES;
  }
  if ((mask & WIDTH_MASK) || mask & (ALL_ATTRIB_MASK)) {
    dp->attributes.spec.line_width    = EASValsGetLineWidth();
    dp->attributes.spec.have_any = YES;
  }
}



static void 
CopyAttributesToActiveSet(GraphicObj *p)
{
  ENURBSS3DWCRec *dp;
  
  dp = (ENURBSS3DWCRec *)p->spec.data;
  
  if (dp->attributes.spec.have_any) {
    EASValsSetFillStyle(dp->attributes.spec.fill_style);
    EASValsSetColor(dp->attributes.spec.color);
    EASValsSetEdgeColor(dp->attributes.spec.edge_color);
    EASValsSetEdgeFlag(dp->attributes.spec.edge_flag);
    EASValsSetShowPoly(dp->attributes.spec.show_polygon);
    EASValsSetShowEntity(dp->attributes.spec.show_surface);
    EASValsSetUTesselIntervals(dp->attributes.spec.tessel_u);
    EASValsSetVTesselIntervals(dp->attributes.spec.tessel_v);
    EASValsSetLineStyle(dp->attributes.spec.line_style);
    EASValsSetLineWidth(dp->attributes.spec.line_width);
  }
}

static BOOLEAN
AttribValAsInAS(GraphicObj *p, unsigned long mask)  
{
  ENURBSS3DWCRec *dp;
  
  dp = (ENURBSS3DWCRec *)p->spec.data;
  
  if ((mask & SHOW_POLY_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.show_polygon != EASValsGetShowPoly())
      return NO;

  if ((mask & SHOW_ENTITY_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.show_surface != EASValsGetShowEntity())
      return NO;
  
  if ((mask & TESSEL_U_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.tessel_u != EASValsGetUTesselIntervals())
      return NO;
  
  if ((mask & TESSEL_V_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.tessel_v != EASValsGetVTesselIntervals())
      return NO;
  
  if ((mask & FILL_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.fill_style != EASValsGetFillStyle())
      return NO;
  
  if ((mask & COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.color != EASValsGetColor())
      return NO;
  
  if ((mask & EDGE_COLOR_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.edge_color != EASValsGetEdgeColor())
      return NO;
  
  if ((mask & EDGE_FLAG_MASK) || (mask & ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.edge_flag != EASValsGetEdgeFlag())
      return NO;

  if ((mask & STYLE_MASK) || mask & (ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.line_style    != EASValsGetLineStyle())
      return NO;
  
  if ((mask & WIDTH_MASK) || mask & (ALL_ATTRIB_MASK)) 
    if (dp->attributes.spec.line_width    != EASValsGetLineWidth())
      return NO;
    
  return YES;
}




static GraphicObj *
DeepCopy(GraphicObj *p)
{
  GraphicObj *gp;
  ENURBSS3DWCRec *thedp;
  caddr_t new_data_p;
  WCRec *points_p, *points_gp;
  float *weights_p, *weights_gp;
	double *knotvector_u_p, *knotvector_u_gp;
	double *knotvector_v_p, *knotvector_v_gp;
	short *multiplicity_u_p, *multiplicity_u_gp;
	short *multiplicity_v_p, *multiplicity_v_gp;
  int i, degree_u, degree_v, pnts_u, pnts_v, spans_u, spans_v, knots_u, knots_v, pnts;
  
  /*	gp = EGDefaultCopyGraphics(p); */
  if ((new_data_p =
       (caddr_t)copy_node(p->spec.data, p->spec.data_length)) == NULL) {
    fprintf(stderr, "Failed copy data for NURBSS3D \n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
    exit(0);
  }
  gp = EGCreateGraphics(p->methods, new_data_p, p->spec.data_length);
  
  thedp = (ENURBSS3DWCRec *)(gp->spec.data);
  degree_u = thedp->degree_u;
  degree_v = thedp->degree_v;
  spans_u = thedp->spans_u;
  spans_v = thedp->spans_v;
  knots_u = thedp->knots_u;
  knots_v = thedp->knots_v;
  pnts_u = thedp->pnts_u;
  pnts_v = thedp->pnts_v;
	//knots_u = pnts_u + degree_u + 1;
	//knots_v = pnts_v + degree_v + 1;

	pnts = pnts_u * pnts_v;
  
  if((thedp -> multiplicity_u = (short *)calloc(spans_u + 1, sizeof(short))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if((thedp -> multiplicity_v = (short *)calloc(spans_v + 1, sizeof(short))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if((thedp -> knotvector_u = (double *)calloc(knots_u, sizeof(double))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if((thedp -> knotvector_v = (double *)calloc(knots_v, sizeof(double))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }

  if((thedp -> points = (WCRec *)calloc(pnts, sizeof(WCRec))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  if((thedp -> weights = (float *)calloc(pnts, sizeof(float))) == NULL){
    fprintf(stderr, "Memory allocation for NURBSS3D in DeepCopy failed\n");
    exit(ESISetErrorExitCode(ELIXIR_ERROR_EXIT));
  }
  
  multiplicity_u_p = ((ENURBSS3DWCRec *)(p->spec.data)) -> multiplicity_u;
  multiplicity_v_p = ((ENURBSS3DWCRec *)(p->spec.data)) -> multiplicity_v;
  knotvector_u_p = ((ENURBSS3DWCRec *)(p->spec.data)) -> knotvector_u;
  knotvector_v_p = ((ENURBSS3DWCRec *)(p->spec.data)) -> knotvector_v;
  points_p = ((ENURBSS3DWCRec *)(p->spec.data)) -> points;
  weights_p = ((ENURBSS3DWCRec *)(p->spec.data)) -> weights;
  multiplicity_u_gp = thedp -> multiplicity_u;
  multiplicity_v_gp = thedp -> multiplicity_v;
  knotvector_u_gp = thedp -> knotvector_u;
  knotvector_v_gp = thedp -> knotvector_v;
  points_gp = thedp -> points;
  weights_gp = thedp -> weights;
  
  for(i = 0; i < pnts; i++){
    points_gp[i].x = points_p[i].x;
    points_gp[i].y = points_p[i].y;
    points_gp[i].z = points_p[i].z;
    weights_gp[i] = weights_p[i];
  }
  return(gp);
}



static void
DeepDestroy(GraphicObj *p)
{
  ENURBSS3DWCRec *thedp;
  thedp = (ENURBSS3DWCRec *)(p->spec.data);
  
  free(thedp->multiplicity_u);
  free(thedp->multiplicity_v);
  free(thedp->knotvector_u);
  free(thedp->knotvector_v);
  free(thedp->points);
  free(thedp->weights);
  
  /* 	EGDefaultDestroyGraphics(p); */
  free_node(thedp);                        /* free specialized data */
  free_node(p);                            /* free top struct */
}

