
#include "Eprimtv.h"

/* there is a problem when non-planar quad is shaded;
	 division to 4 triangles yields sometimes shading that is not smooth enough;
   therefore subdivision of quad in both directions is performed first
	 and then shading based on division to 4 triangles is applied;
	 to make this more adaptive, the area of drawn quad in device coordinate
	 is first compared to AREA;
	 for small enough quads subdivision is vasting the time, therefore
	 standard shading based on division to 4 triangles is used;
	 for larger quads square root of the ratio quad area to AREA should be used
	 to calculate number of subdivisions;
	 to make the subdivision progresive with decreasing size of quad fourth root
	 was actually employed;
	 this concept is a compromise between efficiency and smoothness;

	 original shading (fast but sometimes not very nice) via 4 triangles 
	 can be enforced by setting AREA = 0, MIN_TESSEL = 0, MAX_TESSEL is irrelevant;

	 note: check on quad planarity is not implemented */


/* specify, how large is the area of quad in pixels before subdivision is considered */

#define AREA             64      /* recommended value 64 */

/* specify, how much the quad is subdivided before is split to 4 triangles */

#define MIN_TESSEL        2        /* if AREA == 0 */       /* recommended value 2 */
#define MAX_TESSEL       16        /* if AREA != 0 */       /* recommended value 16 */



#define POINTS_IDENTICAL(P1, P2) \
     ((P1.x == P2.x) && (P1.y == P2.y) && (P1.z == P2.z))

#define SHRINK_VALUES(SHRINK, V1, V2, V3, V4, VC)          \
  {                                                        \
    (V1) = SHRINK * (V1) + (1 - SHRINK) * (VC);            \
    (V2) = SHRINK * (V2) + (1 - SHRINK) * (VC);            \
    (V3) = SHRINK * (V3) + (1 - SHRINK) * (VC);            \
    (V4) = SHRINK * (V4) + (1 - SHRINK) * (VC);            \
  }


#define SHRINK_QUAD_DATA(theSHRINK, fstV, scndV, thrdV, frthV)             \
  {                                                                         \
     double theCV;                                                          \
                                                                            \
     theCV   = (fstV); theCV += (scndV); theCV += (thrdV); theCV += (frthV);\
     theCV   *= 0.25;                                                       \
     (fstV)  = theCV + (theSHRINK) * ((fstV)  - theCV);                     \
     (scndV) = theCV + (theSHRINK) * ((scndV) - theCV);                     \
     (thrdV) = theCV + (theSHRINK) * ((thrdV) - theCV);                     \
     (frthV) = theCV + (theSHRINK) * ((frthV) - theCV);                     \
   }



#define COMPUTE_CENTER(CENTER, P1, P2, P3, P4)  \
  {                                             \
    (CENTER)->x =  (P1)->x;                     \
    (CENTER)->x += (P2)->x;                     \
    (CENTER)->x += (P3)->x;                     \
    (CENTER)->x += (P4)->x;                     \
    (CENTER)->x *= 0.25;                        \
    (CENTER)->y =  (P1)->y;                     \
    (CENTER)->y += (P2)->y;                     \
    (CENTER)->y += (P3)->y;                     \
    (CENTER)->y += (P4)->y;                     \
    (CENTER)->y *= 0.25;                        \
    (CENTER)->z =  (P1)->z;                     \
    (CENTER)->z += (P2)->z;                     \
    (CENTER)->z += (P3)->z;                     \
    (CENTER)->z += (P4)->z;                     \
    (CENTER)->z *= 0.25;                        \
  }


#define COMPUTE_COMBINATION(POINT, P1, P2, P3, P4, XI, ONE_M_XI, ETA, ONE_M_ETA)    \
  {                                                                                 \
		(POINT)->x = (P1)->x*(ONE_M_XI)*(ONE_M_ETA) + (P2)->x*(XI)*(ONE_M_ETA) + (P3)->x*(XI)*(ETA) + (P4)->x*(ONE_M_XI)*(ETA);   \
		(POINT)->y = (P1)->y*(ONE_M_XI)*(ONE_M_ETA) + (P2)->y*(XI)*(ONE_M_ETA) + (P3)->y*(XI)*(ETA) + (P4)->y*(ONE_M_XI)*(ETA);   \
		(POINT)->z = (P1)->z*(ONE_M_XI)*(ONE_M_ETA) + (P2)->z*(XI)*(ONE_M_ETA) + (P3)->z*(XI)*(ETA) + (P4)->z*(ONE_M_XI)*(ETA);   \
  }



void 
EVQuad3DPrimitive(EView        *v_p,
                  GraphicObj   *g_p,
                  WCRec        *first,
                  WCRec        *second,
                  WCRec        *third,
                  WCRec        *fourth,
                  EPixel        pixel,
                  int           fill,
                  int           edge_flag,
                  EPixel        edge_pixel,
                  float         shrink,
                  EDrawMode     draw_mode)
{
  WCRec s1, s2, s3, s4, center;
  int how_many_in = 0;
  EV_line_end_point_rec ap;
	
  COPY_WCREC(&s1, first);
  COPY_WCREC(&s2, second);
  COPY_WCREC(&s3, third);
  COPY_WCREC(&s4, fourth);
  if (shrink != 1) {
    SHRINK_QUAD_DATA(shrink, s1.x, s2.x, s3.x, s4.x);
    SHRINK_QUAD_DATA(shrink, s1.y, s2.y, s3.y, s4.y);
    SHRINK_QUAD_DATA(shrink, s1.z, s2.z, s3.z, s4.z);
  }
	
  if (v_p->should_clip_by_model_planes  && EGIsClippable(g_p)) {
    if (v_p->model_clip_plane.clip_is_on) {
      if (EGInvisibleWhenClipped(g_p)) {
        COPY_WCREC(&ap.loc.wc, &s1);
        if (point_is_in_IN_halfspace(&ap, &v_p->model_clip_plane.center,
                                     &v_p->model_clip_plane.normal))
          how_many_in++;
        COPY_WCREC(&ap.loc.wc, &s2);
        if (point_is_in_IN_halfspace(&ap, &v_p->model_clip_plane.center,
                                     &v_p->model_clip_plane.normal))
          how_many_in++;
        COPY_WCREC(&ap.loc.wc, &s3);
        if (point_is_in_IN_halfspace(&ap, &v_p->model_clip_plane.center,
                                     &v_p->model_clip_plane.normal))
          how_many_in++;
        COPY_WCREC(&ap.loc.wc, &s4);
        if (point_is_in_IN_halfspace(&ap, &v_p->model_clip_plane.center,
                                     &v_p->model_clip_plane.normal))
          how_many_in++;
        if (how_many_in == 0)
          return;
        if (how_many_in != 4)
          return;
      }
    }
  }
	
  if (v_p->render_mode == WIRE_RENDERING || fill == FILL_HOLLOW) {
    if (draw_mode == DRAW_MODE_XORDRAW
        && POINTS_IDENTICAL(s3, s4) && POINTS_IDENTICAL(s3, s2))
      EVLine3DPrimitive(v_p, g_p, &s1, &s2, SOLID_STYLE, pixel, 0, 1.0, draw_mode);
    else {
			EV_render_mode = v_p->render_mode;
			if(EV_render_mode == FILLED_HIDDEN_RENDERING || EV_render_mode == CONST_SHADING_RENDERING){
				if(edge_flag){
					EVLine3DPrimitive(v_p, g_p, &s1, &s2, SOLID_STYLE, edge_pixel, 0, 1.0, draw_mode);
					EVLine3DPrimitive(v_p, g_p, &s2, &s3, SOLID_STYLE, edge_pixel, 0, 1.0, draw_mode);
					EVLine3DPrimitive(v_p, g_p, &s3, &s4, SOLID_STYLE, edge_pixel, 0, 1.0, draw_mode);
					EVLine3DPrimitive(v_p, g_p, &s4, &s1, SOLID_STYLE, edge_pixel, 0, 1.0, draw_mode);
				}
			}
			else{
				EVLine3DPrimitive(v_p, g_p, &s1, &s2, SOLID_STYLE, pixel, 0, 1.0, draw_mode);
				EVLine3DPrimitive(v_p, g_p, &s2, &s3, SOLID_STYLE, pixel, 0, 1.0, draw_mode);
				EVLine3DPrimitive(v_p, g_p, &s3, &s4, SOLID_STYLE, pixel, 0, 1.0, draw_mode);
				EVLine3DPrimitive(v_p, g_p, &s4, &s1, SOLID_STYLE, pixel, 0, 1.0, draw_mode);
			}
    }
  } else {
		float s = 1;
		int tessel = MIN_TESSEL;

		/* it may be usefull to check whether quad is planar;
			 if yes tessel can be safely set to zero */

		if(tessel != 0 && v_p->render_mode == CONST_SHADING_RENDERING && AREA != 0){
			int DCx1, DCx2, DCx3, DCx4;
			int DCy1, DCy2, DCy3, DCy4;
			int dx1, dy1, dx2, dy2;
			int area;
			
			EVWCtoDC(v_p, &s1, &DCx1, &DCy1);
			EVWCtoDC(v_p, &s2, &DCx2, &DCy2);
			EVWCtoDC(v_p, &s3, &DCx3, &DCy3);
			EVWCtoDC(v_p, &s4, &DCx4, &DCy4);

			dx1 = DCx2 - DCx1;
			dy1 = DCy2 - DCy1;
			dx2 = DCx4 - DCx1;
			dy2 = DCy4 - DCy1;
			
			area = dx1 * dy2 - dx2 * dy1;
			
			dx1 = DCx4 - DCx3;
			dy1 = DCy4 - DCy3;
			dx2 = DCx2 - DCx3;
			dy2 = DCy2 - DCy3;
			
			area += dx1 * dy2 - dx2 * dy1;
			
			area = abs(area);
			if(area <= AREA)
				tessel = 0;
			else{
				tessel = (int)((sqrt((double)area / AREA))) + 2;
				if(tessel > MAX_TESSEL)tessel = MAX_TESSEL;
			}
		}
		
		if(tessel != 0 && v_p->render_mode == CONST_SHADING_RENDERING){
			WCRec p1, p2, p3, p4;
			double xi, eta, dxi, deta;
			int i, j;
			
			dxi = deta = 1.0 / tessel;
			
			for(xi = 0.0, i = 0; i < tessel; i++, xi += dxi){
				COMPUTE_COMBINATION(&p1, &s1, &s2, &s3, &s4, xi, 1.0 - xi, 0.0, 1.0);
				COMPUTE_COMBINATION(&p2, &s1, &s2, &s3, &s4, xi + dxi, 1.0 - xi - dxi, 0.0, 1.0);
				
				for(eta = deta, j = 0; j < tessel; j++, eta += deta){
					COMPUTE_COMBINATION(&p3, &s1, &s2, &s3, &s4, xi + dxi, 1.0 - xi - dxi, eta, 1.0 - eta);
					COMPUTE_COMBINATION(&p4, &s1, &s2, &s3, &s4, xi, 1.0 - xi, eta, 1.0 - eta);
					
					COMPUTE_CENTER(&center, &p1, &p2, &p3, &p4);
					EVTriangle3DPrimitive(v_p, g_p, &p1, &p2, &center,
																pixel, fill,
																NO, edge_pixel, s, draw_mode);
					EVTriangle3DPrimitive(v_p, g_p, &p2, &p3, &center,
																pixel, fill,
																NO, edge_pixel, s, draw_mode);
					EVTriangle3DPrimitive(v_p, g_p, &p3, &p4, &center,
																pixel, fill,
																NO, edge_pixel, s, draw_mode);
					EVTriangle3DPrimitive(v_p, g_p, &p4, &p1, &center,
																pixel, fill,
																NO, edge_pixel, s, draw_mode);
					
					COPY_WCREC(&p1, &p4);
					COPY_WCREC(&p2, &p3);
				}
			}
		}
		else{
			COMPUTE_CENTER(&center, &s1, &s2, &s3, &s4);
			EVTriangle3DPrimitive(v_p, g_p, &s1, &s2, &center,
														pixel, fill,
														NO, edge_pixel, s, draw_mode);
			EVTriangle3DPrimitive(v_p, g_p, &s2, &s3, &center,
														pixel, fill,
														NO, edge_pixel, s, draw_mode);
			EVTriangle3DPrimitive(v_p, g_p, &s3, &s4, &center,
														pixel, fill,
														NO, edge_pixel, s, draw_mode);
			EVTriangle3DPrimitive(v_p, g_p, &s4, &s1, &center,
														pixel, fill,
														NO, edge_pixel, s, draw_mode);
		}

    if (edge_flag) {
      int ls = SOLID_STYLE;
      double dn = DELTA_N();
      WCRec n, p1, p2, p3, p4;

#define INCR(p, dn, n) { p.x += dn * n.x; p.y += dn * n.y; p.z += dn * n.z; }

      EVGetNormal(v_p, &n);
      COPY_WCREC(&p1, &s1); INCR(p1, dn, n);
      COPY_WCREC(&p2, &s2); INCR(p2, dn, n);
      COPY_WCREC(&p3, &s3); INCR(p3, dn, n);
      COPY_WCREC(&p4, &s4); INCR(p4, dn, n);

      EVLine3DPrimitive(v_p, g_p, &p1, &p2, ls, edge_pixel, 0, 1.0, draw_mode);
      EVLine3DPrimitive(v_p, g_p, &p2, &p3, ls, edge_pixel, 0, 1.0, draw_mode);
      EVLine3DPrimitive(v_p, g_p, &p3, &p4, ls, edge_pixel, 0, 1.0, draw_mode);
      EVLine3DPrimitive(v_p, g_p, &p4, &p1, ls, edge_pixel, 0, 1.0, draw_mode);
    }
  }
}
